/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** A client that attempts to create paths */

#include <iostream>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "../udp-lib/UDPLib.h"
#include "../serv-comp/ServCompIFPkt.h"
#include "../serv-comp/ServComp.h"
#include "../util/LogLine.h"
#include "../util/Util.h"

#define CLIENT_PORT 14000

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <sc-host>\n", progname);
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  // check command line args
  if(argc != 2) { usage(argv[0]); exit(1); }
  const char *scHost = argv[1];
  struct in_addr scIPAddr = nslookup(scHost); // DNS lookup

  // prepare the UDP socket for listening
  int udp_sock = sock_and_bind(CLIENT_PORT);
  if(udp_sock < 0) { perror("sock_and_bind"); exit(3); }

  // init the udp library
  UDPLib::init();

  // prepare the packet to send
  UDPLibPkt_t uPkt;
  ServCompIFPkt_t sPkt;
  sPkt.fixed.type = APP_CREATE_PATH;
  sPkt.fixed.app_seq = 1;
  sPkt.fixed.port = CLIENT_PORT;
  sPkt.addPair(PREV_DEST_SPEC_F, "dummy_dest_spec");
  sPkt.addPair(SERVICES_F, "s0 s1 s3");
  sPkt.fixed.path_origin_scid = 8;
  int uPktLen = sPkt.toStr(uPkt.data, sizeof(uPkt.data));

  // do a synchronous request response
  UDPLibPkt_t resp_uPkt;
  ServCompIFPkt_t resp_sPkt;
  int resp_len;
  bool use_tunnel = false; // should not send pkt through the delaying tunnel
  unsigned long count = 0, rcv_count = 0;
  int send_fd = udp_sock; // can use same socket for sending
  unsigned short dest_scid = LogLine::INVALID_SCID;
  unsigned short my_scid = LogLine::INVALID_SCID;
  int res = UDPLib::syncReqResp(send_fd, &uPkt, uPktLen, scIPAddr,
				SERV_COMP_CLIENT_PORT, udp_sock, &resp_uPkt,
				&resp_len, NULL, use_tunnel, &count,
				&rcv_count, dest_scid, my_scid);
  if(res != 1) {
    cerr << "res = " << res << endl;
    exit(3);
  }

  int parseRes = resp_sPkt.parsePkt(resp_uPkt.data, resp_len);
  if(!parseRes) {
    cerr << "parseRes = " << parseRes << endl;
    exit(3);
  }

  cerr << "success = \t" << resp_sPkt.fixed.success << endl
       << "app_seq = \t" << resp_sPkt.fixed.app_seq << endl
       << "pathID = \t" << resp_sPkt.fixed.pathID << endl
       << "path_dest_scid = \t" << resp_sPkt.fixed.path_dest_scid << endl;

} // End main()
