/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements a simple service */

#include <iostream>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "../udp-lib/UDPLib.h"
#include "../serv-comp/ServCompIFPkt.h"
#include "../serv-comp/ServComp.h"

#define SERVICE_PORT 24000

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s\n", progname);
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 1) { usage(argv[0]); exit(1); }

  // prepare the UDP socket for listening
  int udp_sock = sock_and_bind(SERVICE_PORT);
  if(udp_sock < 0) { perror("socket"); exit(3); }

  // init the udp library
  UDPLib::init();

  cerr << "service running on port " << SERVICE_PORT << "\a" << endl;

  // respond to all createServ requests, forever
  for(;;) {
    // recv the pkt
    UDPLibPkt_t uPkt;
    ServCompIFPkt_t sPkt;
    int recv_len;
    struct sockaddr_in from_addr;
    unsigned long rcv_count = 0;
    bool pkt_ready = false;
    int res = UDPLib::blockingRecv(udp_sock, &uPkt, &recv_len,
				   &from_addr, NULL, &rcv_count, pkt_ready);
    if(res != 1) {
      cerr << "res = " << res << endl;
      continue;
    }

    // parse it
    int parseRes = sPkt.parsePkt(uPkt.data, recv_len);
    if(parseRes != 1) {
      cerr << "parseRes = " << parseRes << endl;
      continue;
    }

    // print contents
    if(gdbg) {
      cerr << "--------------------------------------------------" << endl
	   << "pkt from " << inet_ntoa(from_addr.sin_addr) << endl
	   << "type = \t" << sPkt.fixed.type << endl
	   << "pathID = \t" << sPkt.fixed.pathID << endl
	   << "path_dest_scid = \t" << sPkt.fixed.path_dest_scid << endl
	   << "prev_dest_spec = \t" << sPkt.getValue(PREV_DEST_SPEC_F) << endl;
    }

    // send response
    ServCompIFPkt_t resp_sPkt;
    resp_sPkt.addPair(PREV_DEST_SPEC_F, "dummy_service_dest_spec");
    resp_sPkt.fixed.success = 1;
    int send_len = resp_sPkt.toStr(uPkt.data, sizeof(uPkt.data));
    int send_fd = udp_sock; // can use same sock to send pkts too
    bool use_tunnel = false; // should not send pkt via the delaying tunnel
    unsigned long count = 0;
    res = UDPLib::sendResp(send_fd, &uPkt, send_len,
			   from_addr.sin_addr, SERV_COMP_SERVER_PORT,
			   use_tunnel, &count);
    if(res != 1) {
      cerr << "res = " << res << endl;
      continue;
    }

  } // End for(;;)

} // End main()
