/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to create lots of paths */

#include <iostream>
#include <fstream>
#include <vector>
#include <queue>
#include <list>

#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <math.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "../udp-lib/UDPLib.h"
#include "../serv-comp/ServCompIFPkt.h"
#include "../serv-comp/ServComp.h"
#include "../util/LogLine.h"
#include "../util/Util.h"
#include "../util/JString.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"

#define CLIENT_PORT 14000

struct meas {
  int pathID;
  unsigned short path_dest_scid;
  JString services;
  unsigned short path_origin_scid;
  int reqNo;
  struct timeval startTime;
  unsigned long creation_time;
  bool res_succ;
  bool success;
  meas() {
    reqNo = -1;
    res_succ = success = false;
  }
};
typedef struct meas meas_t;

meas_t *M;
int numPaths;
vector<sc_node_t> sc_list;

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <scid-mapping-file> <num-paths>\n", progname);
} // End usage()

struct cp_req {
  int reqNo;
  int ind;
};
typedef struct cp_req cp_req_t;

queue<cp_req_t, list<cp_req_t> > reqs;
int numSent = 0;
reqRespBlock_t *rrBlock;
UDPPktBlock_t *pktBlock;
int udp_sock;
Uniform U;
ServCompIFPkt_t sPkt;

void sendReq() {
  bool use_tunnel = false; // should not send pkt through the delaying tunnel
  // dummy, since not using tunnel
  unsigned short dest_scid = LogLine::INVALID_SCID;
  // also, dummy for the same reason
  unsigned short my_scid = LogLine::INVALID_SCID;
  int send_fd = udp_sock; // can use same socket for sending

  if(!cfg_read_paths) {
    int s0 = (int)floor(U.Next()*9.999);
    int s1 = (int)floor(U.Next()*9.999);
    while(s1 == s0) {
      s1 = (int)floor(U.Next()*9.999);
    }
    char serv_str[128];
    sprintf(serv_str, "s%d s%d", s0, s1);
    M[numSent].services = serv_str;
  }

  sPkt.addPair(SERVICES_F, M[numSent].services.cstr());

  if(!cfg_read_paths) { M[numSent].pathID = numSent; }
  sPkt.fixed.app_seq = M[numSent].pathID;

  UDPLibPkt_t *uPkt = pktBlock->getPkt();
  if(uPkt == NULL) {
    fprintf(stderr, "No space for uPkt\n");
    exit(33);
  }
  int uPktLen = sPkt.toStr(uPkt->data, sizeof(uPkt->data));

  int sc_index = -1;
  if(!cfg_read_paths) {
    // get a random sc node to send request to
    sc_index = (int)floor(U.Next()*(sc_list.size()-0.01));
    M[numSent].path_dest_scid = sc_list[sc_index].scid;
  } else {
    for(unsigned int i = 0; i < sc_list.size(); i++) {
      if(M[numSent].path_dest_scid == sc_list[i].scid) {
	sc_index = i; break;
      }
    }
  }
  if(sc_index < 0) {
    cerr << "sc_index=" << sc_index << endl;
    exit(87);
  }

  // get space for ack packet
  UDPLibPkt_t *ackPkt = pktBlock->getPkt();
  if(ackPkt == NULL) {
    fprintf(stderr, "No space for ackPkt\n");
    exit(33);
  }

  gettimeofday(&M[numSent].startTime, NULL);

  struct timeval timeout = { 50, 0 }; // 50sec
  cp_req_t r1;
  r1.ind = numSent;
  r1.reqNo = rrBlock->insertRequest
    (send_fd, uPkt, uPktLen, sc_list[sc_index].ipAddr, SERV_COMP_CLIENT_PORT,
     udp_sock, ackPkt, timeout, use_tunnel, dest_scid, my_scid);
  if(r1.reqNo < 0) {
    fprintf(stderr, "no space to insert request\n");
    exit(66);
  }

  reqs.push(r1);
  numSent++;
  fputc('+', stderr);

} // End sendReq()

void checkResp() {
  for(int i = 0; i < 30; i++) {
    if(reqs.empty()) break;
    cp_req_t r1 = reqs.front();
    reqs.pop();
    int res = rrBlock->checkResp(r1.reqNo, NULL, NULL);

    if(res != 0) {

      M[r1.ind].res_succ = (res == 1);
      M[r1.ind].creation_time = (unsigned long)now(M[r1.ind].startTime);
      if(M[r1.ind].res_succ) {
	fputc('.', stderr);

	ServCompIFPkt_t resp_sPkt;
	int parseRes = resp_sPkt.parsePkt
	  (rrBlock->reqs[r1.reqNo].recv_pkt->data,
	   rrBlock->reqs[r1.reqNo].recv_len);
	if(!parseRes) {
	  cerr << "parseRes = " << parseRes << endl;
	  exit(3);
	}
	int succ = resp_sPkt.fixed.success;
	M[r1.ind].success = (succ == 1);
      } else {
	fputc('#', stderr);
      }

      // ack has been received, or timeout has happened
      // free the send and receive packets
      pktBlock->freePkt(rrBlock->reqs[r1.reqNo].send_pkt);
      pktBlock->freePkt(rrBlock->reqs[r1.reqNo].recv_pkt);
      // free the udp-lib entry
      rrBlock->freeEntry(r1.reqNo);

    } else {
      reqs.push(r1);
    }

    } // End for checking for acks/timeouts
} // End checkResp()

void readPaths() {
  ifstream fin("../serv-comp/paths.cfg");
  if(fin.fail()) {
    cerr << "Can't open ../serv-comp/paths.cfg for reading" << endl;
    exit(43);
  }

  int line_num = 0;
  for(;;) {
    // read each line
    if(fin.eof()) break;
    char line[2048];
    fin.getline(line, sizeof(line)); // read a line
    if(fin.eof()) break;
    if(fin.fail()) {
      cerr << "Error reading" << endl; exit(99);
    }
    line_num++;
    LogLine_t ll(line);
    if(!ll.succ) {
      cerr << "Format wrong while reading line: " << line_num << endl;
      cerr << line << endl;
      exit(9);
    }

    if((ll.pathID < 0) || (ll.pathID >= numPaths)) {
      cerr << "pathID=" << ll.pathID << ", numPaths=" << numPaths << endl;
      exit(98);
    }

    meas_t &m1 = M[ll.pathID];
    m1.pathID = ll.pathID;
    m1.path_origin_scid = ll.path_origin_scid;
    m1.path_dest_scid = ll.path_dest_scid;
    m1.services = ll.strip_getValue("services");

  } // End for(;;) reading file

  fin.close();
} // End readPaths()

int main(int argc, char *argv[]) {
  util_init();
  // check command line args
  if(argc != 3) { usage(argv[0]); exit(1); }
  const char *scid_mapping = argv[1];
  numPaths = atoi(argv[2]);

  // alloc for measurements
  M = new meas_t[numPaths];
  if(M == NULL) {
    cerr << "Can't alloc for measurements" << endl;
    exit(4);
  }

  // read the scid mapping file
  util_readSCIDMapping(scid_mapping, sc_list);

  // read the paths file
  if(cfg_read_paths) { readPaths(); }

  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long seedLong = tv.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  fprintf(stderr, "setting random seed done with %ld(%f)\n",
	  seedLong, seedFloat);
  Uniform::Set(seedFloat);

  // prepare the UDP socket for listening
  udp_sock = sock_and_bind(CLIENT_PORT);
  if(udp_sock < 0) { perror("sock_and_bind"); exit(3); }

  // init the udp library
  UDPLib::init();

  // get the request response block for sends
  rrBlock = UDPLib::getReqRespBlock();
  if(rrBlock == NULL) { cerr << "rrBlock is NULL" << endl; exit(23); }

  // get a block of packet space for sends
  pktBlock = UDPPktBlock::getUDPPktBlock();
  if(pktBlock == NULL) { cerr << "pktBlock is NULL" << endl; exit(24); }

  // prepare the packet to send
  sPkt.fixed.type = APP_CREATE_PATH;
  sPkt.fixed.port = CLIENT_PORT;
  sPkt.addPair(PREV_DEST_SPEC_F, "dummy_dest_spec");
  //sPkt.fixed.path_origin_scid = 8;

  // do a synchronous request response

  // create the paths
  while((numSent < numPaths) || (!reqs.empty())) {
    if((numSent < numPaths) && (reqs.size() <= 100)) {
      sendReq();
      if(!cfg_read_paths) { usleep(30*1000); }
    } else {
      usleep(30*1000);
    }
    if(((numSent%100) == 0) && (numSent < numPaths)) {
      fprintf(stderr, "(numSent=%d)", (int)numSent);
    }

    // check for response for a bunch of requests
    checkResp();
  } // End while creating paths and waiting for response

  fputc('\n', stderr);

  // print some stats
  int numSucc = 0, maxIndex = 0;
  float sumLat = 0, sumSqLat = 0;
  for(int i = 0; i < numPaths; i++) {
    if(M[i].res_succ && M[i].success) {
      sumLat += M[i].creation_time;
      sumSqLat += (M[i].creation_time * M[i].creation_time);
      numSucc++;
      if(M[maxIndex].creation_time < M[i].creation_time) {
	maxIndex = i;
      }
    } else {
      cerr << "Failure at " << M[i].path_dest_scid << endl;
    }
  }

  float avg = 0, se = 0;
  if(numSucc > 0) {
    avg = sumLat/numSucc;
    se = sqrt(sumSqLat/numSucc - avg*avg);
  }

  cerr << '\a' << endl;
  cerr << "Total number of paths = \t" << numPaths << endl;
  cerr << "Created successfully  = \t" << numSucc << " ("
       << (100.0*numSucc/numPaths) << "%)" << endl;
  cerr << "Avg creation time     = \t" << avg << endl;
  cerr << "Standard error        = \t" << se << endl;
  cerr << "Max creation time     = \t" << M[maxIndex].creation_time << endl;
  cerr << "Max index             = \t" << maxIndex << endl;

} // End main()
