/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements generic_client_info.h */

#include <unistd.h>

#include "generic_client_info.h"
#include "../util/Util.h"
#include "../udp-lib/UDPLib.h"

reqRespBlock_t *generic_client_info::rrBlock;
struct in_addr generic_client_info::myIP;
Debug_t *generic_client_info::dbg = NULL; // for debugging
char generic_client_info::name[128]; // same as the one in the service structure

unsigned int generic_client_info::PATH_HB_PERIOD = 1*1000; // 1sec
unsigned int generic_client_info::PATH_HB_TIMEOUT = 6*1000; // 6sec

void generic_client_info::parseUpstreamInfo(ServCompIFPkt_t *sPkt,
					    ServCompIFPkt_t *resp_sPkt) {
  if(!parseIPPort(sPkt->getValue(UPSTREAM_SPEC_F), &upstreamIP,
		  &upstreamPort)) {
    resp_sPkt->fixed.success = 0; return;
  }
  serv_upstream_scid = sPkt->fixed.serv_upstream_scid;
  resp_sPkt->fixed.success = 1;
} // End generic_client_info::parseUpstreamInfo()

void generic_client_info::sendUpstreamInfo() {
  ServCompIFPkt_t req_sPkt;
  req_sPkt.fixed.type = APP_UPSTREAM_INFO;
  req_sPkt.fixed.pathID = pathID;
  req_sPkt.fixed.path_dest_scid = path_dest_scid;
  req_sPkt.fixed.version = version;
  JString upstream_spec = ip2str(myIP) + ":" + long2str(data_listen_port);
  req_sPkt.addPair(UPSTREAM_SPEC_F, upstream_spec.cstr());
  req_sPkt.fixed.serv_upstream_scid = serv_my_scid; // i'm upstream of
  // who i'm sending to
  int send_len = req_sPkt.toStr(ctl_reqPkt.data, sizeof(ctl_reqPkt.data));
  struct timeval timeout = { 2, 0 }; // 2sec

  // send the request upstream to get text
  int send_fd = ctl_sock_fd; // use same socket for sending
  // send pkt through the delaying tunnel, if I'm not the last in path
  bool use_tunnel = ((serv_dest_scid != serv_my_scid) &&
		     (serv_dest_scid != LogLine::INVALID_SCID));
  ctl_reqNo = rrBlock->insertRequest
    (send_fd, &ctl_reqPkt, send_len, destIP, destPort+CTL_OFFSET,
     ctl_ack_sock_fd, &ctl_ackPkt, timeout, use_tunnel,
     serv_dest_scid, serv_my_scid);
  if(ctl_reqNo == -1) {
    cerr << "generic_cl.sendUpstreamInfo: no space for request\n";
    exit(74);
  }

  dbg->print() << name << " send-upstream-info p_id=" << pathID << ':'
	       << path_dest_scid << " serv_dest_scid=" << serv_dest_scid
	       << " path_ver=" << version << " serv_my_scid="
	       << serv_my_scid << " dest=" << ip2str(destIP) << ':'
	       << (destPort+CTL_OFFSET) << endl;
} // End generic_client_info::sendUpstreamInfo()

void generic_client_info::handleCtl() {
  // recv the pkt
  UDPLibPkt_t uPkt; ServCompIFPkt_t sPkt;
  int recv_len; struct sockaddr_in from_addr;
  unsigned long rcv_count = 0; bool pkt_ready = true;
  int res = UDPLib::blockingRecv(ctl_sock_fd, &uPkt, &recv_len,
				 &from_addr, NULL, &rcv_count, pkt_ready);
  if(res != 1) { cerr << "res = " << res << endl; return; }
  // parse it
  int parseRes = sPkt.parsePkt(uPkt.data, recv_len);
  if(parseRes != 1) { cerr << "parseRes = " << parseRes << endl; return; }

  ServCompIFPkt_t resp_sPkt;
  // get upstream information
  if(sPkt.fixed.type == APP_UPSTREAM_INFO) {
    parseUpstreamInfo(&sPkt, &resp_sPkt);
    dbg->print() << name << " recv-upstream-info p_id=" << pathID << ':'
		 << path_dest_scid << " path_ver=" << version
		 << " serv_my_scid=" << serv_my_scid
		 << " serv_upstream_scid=" << serv_upstream_scid << endl;
    sendUpstreamInfo(); // send my upstream to the elt downstream of me
  } else if(sPkt.fixed.type == APP_HB) {
    //dbg->print() << name << " recv-hb p_id=" << pathID << ':'
    //<< path_dest_scid << " path_ver=" << version << endl;
    numHBRecd++; gettimeofday(&last_hb_recv_time, NULL);
    return; // no need to respond (downstream)
  } else if(sPkt.fixed.type == APP_KILL) {
    dbg->print() << name << " path-kill p_id=" << pathID << ':'
		 << path_dest_scid << " path_ver=" << version << endl;
    if(upstreamPort > 0) sendHB(APP_KILL); // send APP_KILL further upstream
    endSession();
    return; // no need to respond (downstream)
  } else {
    cerr << "cl.handleCtl: unknown type ["
	 << ServCompIFPkt::type2str((ServCompPktType_t)sPkt.fixed.type)
	 << "]\n";
    resp_sPkt.fixed.success = 0;
  }

  // send response
  int send_len = resp_sPkt.toStr(uPkt.data, sizeof(uPkt.data));
  int send_fd = ctl_ack_sock_fd; // can use same sock to send pkts too
  // should send pkt via the delaying tunnel
  bool use_tunnel = ((serv_upstream_scid != serv_my_scid) &&
		     (serv_my_scid != LogLine::INVALID_SCID));
  unsigned long count = 0;
  res = UDPLib::sendResp(send_fd, &uPkt, send_len,
			 from_addr.sin_addr, upstreamPort+CTL_ACK_OFFSET,
			 use_tunnel, &count);
  if(res != 1) {
    cerr << "cl.handleCtl: res = " << res << endl; return;
  }

} // End generic_client_info::handleCtl()

void generic_client_info::handleCtlAck() {
  int ctl_reqNo_recd = rrBlock->check_one(ctl_ack_sock_fd);
  if(ctl_reqNo_recd < 0) { cerr << "[g:r=-1]"; return; }
  if(ctl_reqNo != ctl_reqNo_recd) {
    cerr << "gs_cl.handleCtlAck: ctl_reqNo mismatch, ctl_reqNo=" << ctl_reqNo
	 << " ctl_reqNo_recd=" << ctl_reqNo_recd << endl;
    exit(99);
  }
  rrBlock->freeEntry(ctl_reqNo); ctl_reqNo = -1;
  dbg->print() << name << " ctl-ack p_id=" << pathID << ':'
	       << path_dest_scid << " path_ver=" << version << endl;
} // End generic_client_info::handleCtlAck()

void generic_client_info::checkCtlSend() {
  // this checks for upstream-info resend
  if(state == IDLE_CL) { return; }
  if(ctl_reqNo == -1) { return; }

  // check if data req has to be resent
  int ackRecd = rrBlock->check_t_o_resend(ctl_reqNo);
  if(ackRecd == 1) {
    rrBlock->freeEntry(ctl_reqNo);
    ctl_reqNo = -1;
  } else if(ackRecd == -1) {
    rrBlock->freeEntry(ctl_reqNo);
    ctl_reqNo = -1;
    endSession();
  } else {
    // still waiting for ack
  }
} // End generic_client_info::checkCtlSend()

void generic_client_info::checkHBSend() {
  if(state == IDLE_CL) return; // why am i being called?
  if(upstreamPort <= 0) return; // no upstream info yet

  bool to_send_now = false;
  if(numHBSent == 0) { to_send_now = true; }
  else {
    float time_since_last_hb_send = now(last_hb_send_time);
    if(time_since_last_hb_send > PATH_HB_PERIOD) {
      to_send_now = true;
    }
  }

  if(!to_send_now) return; // not yet time to send next HB

  sendHB(APP_HB);
  //dbg->print() << name << " send-hb p_id=" << pathID << ':'
  //<< path_dest_scid << " path_ver=" << version << endl;
  gettimeofday(&last_hb_send_time, NULL);
  numHBSent++;
} // End generic_client_info::checkHBSend()

void generic_client_info::sendHB(ServCompPktType_t type) {
  UDPLibPkt_t uPkt; ServCompIFPkt_t req_sPkt;
  req_sPkt.fixed.type = type;
  req_sPkt.fixed.pathID = pathID;
  req_sPkt.fixed.path_dest_scid = path_dest_scid;
  req_sPkt.fixed.version = version;
  int upkt_len = req_sPkt.toStr(uPkt.data, sizeof(uPkt.data));
  int send_len = upkt_len + sizeof(uPkt) - MAXDATASZ;

  struct sockaddr_in to_addr;
  bzero((char *)&to_addr, sizeof(to_addr));
  to_addr.sin_family = PF_INET;
  to_addr.sin_addr = upstreamIP;
  to_addr.sin_port = htons(upstreamPort+CTL_OFFSET);

  bool use_tunnel = ((serv_my_scid != serv_upstream_scid) &&
		     (serv_my_scid != LogLine::INVALID_SCID));
  unsigned long numSentDummy;
  int res = lib_sendto(ctl_sock_fd, &uPkt, send_len, 0, &to_addr,
		       use_tunnel, &numSentDummy, serv_upstream_scid,
		       serv_my_scid);
  if(res != send_len) {
    if(res < 0) { perror("cl.checkHBSend.lib_sendto"); }
    fprintf(stderr, "res=%d, send_len=%d\n", res, send_len);
  }

  dbg->print() << name << " send-hb type=" << ServCompIFPkt::type2str(type)
	       << " p_id=" << pathID << ':' << path_dest_scid
	       << " path_ver=" << version << endl;
} // End generic_client_info::sendHB()

void generic_client_info::checkPathTimeout() {
  float time_since_last_hb_recv = now(last_hb_recv_time);
  if(time_since_last_hb_recv > PATH_HB_TIMEOUT) {
    dbg->print() << name << " path-timeout p_id=" << pathID << ':'
		 << path_dest_scid << " path_ver=" << version << endl;
    endSession();
  }
} // End generic_client_info::checkPathTimeout()

bool generic_client_info::parseDestSpec(const char *dest_ip_port) {
  return parseIPPort(dest_ip_port, &destIP, &destPort);
} // End generic_client_info::parseDestSpec()

bool generic_client_info::parseIPPort(const char *dest_ip_port,
				      struct in_addr *ip, int *port) {
  if(dest_ip_port == NULL) { return false; }
  char *col_pos = index(dest_ip_port, ':');
  if(col_pos == NULL) {
    cerr << "generic: dest spec format error [" << dest_ip_port << "]\n";
    return false;
  }
  *col_pos = 0;
  inet_aton(dest_ip_port, ip);
  *port = atoi(col_pos+1);
  return true;
} // End generic_client_info::parseIPPort()

void generic_client_info::endSession() {
  close(data_sock_fd); close(ctl_sock_fd);
  close(data_ack_sock_fd); close(ctl_ack_sock_fd); close(state_sock_fd);
  if(reqNo != -1) {
    rrBlock->freeEntry(reqNo); reqNo = -1;
  }
  if(ctl_reqNo != -1) {
    rrBlock->freeEntry(ctl_reqNo); ctl_reqNo = -1;
  }
  init_fields(); // init all of them for the next client
} // End generic_client_info::endSession()

void generic_client_info::init_fields() {
  pathID = -1; path_dest_scid = LogLine::INVALID_SCID; version = 0;
  state = IDLE_CL;
  data_sock_fd = ctl_sock_fd = data_ack_sock_fd =
    ctl_ack_sock_fd = state_sock_fd = -1;
  reqNo = ctl_reqNo = -1;
  serv_dest_scid = serv_my_scid =
    serv_upstream_scid = LogLine::INVALID_SCID;
  numHBSent = numHBRecd = 0;
  data_listen_port = ctl_listen_port = data_ack_listen_port =
    ctl_ack_listen_port = state_listen_port = 0;
  destPort = upstreamPort = 0;
} // End generic_client_info::init_fields()

void generic_client_info::init_from_sc(ServCompIFPkt_t *sPkt,
				       ServCompIFPkt_t *resp_sPkt,
				       int arg_data_listen_port) {
  init_fields();

  pathID = sPkt->fixed.pathID;
  path_dest_scid = sPkt->fixed.path_dest_scid;
  version = sPkt->fixed.version;
  if(!parseDestSpec(sPkt->getValue(PREV_DEST_SPEC_F))) {
    resp_sPkt->fixed.success = 0; return;
  }

  serv_dest_scid = sPkt->fixed.serv_dest_scid;
  serv_my_scid = sPkt->fixed.serv_my_scid;
  // note: NOT initializing serv_upstream_scid

  if(!init_sock(arg_data_listen_port)) {
    close(data_sock_fd); close(data_ack_sock_fd);
    close(ctl_sock_fd); close(ctl_ack_sock_fd); close(state_sock_fd);
    resp_sPkt->fixed.success = 0; return;
  }

  // pretend as if you have recd a HB
  gettimeofday(&last_hb_recv_time, NULL);
} // End generic_client_info::init_from_sc()

bool generic_client_info::init_sock(int arg_data_listen_port) {
  // init socket stuff for the client
  // (1) data port
  data_listen_port = arg_data_listen_port + DATA_OFFSET;
  data_sock_fd = sock_and_bind(data_listen_port);
  if(data_sock_fd < 0) {
    perror("generic.data:sock_and_bind"); return false;
  }
  // (2) ctl port
  ctl_listen_port = arg_data_listen_port + CTL_OFFSET;
  ctl_sock_fd = sock_and_bind(ctl_listen_port);
  if(ctl_sock_fd < 0) {
    perror("generic.ctl:sock_and_bind"); return false;
  }
  // (3) data ack port
  data_ack_listen_port = arg_data_listen_port + DATA_ACK_OFFSET;
  data_ack_sock_fd = sock_and_bind(data_ack_listen_port);
  if(data_ack_sock_fd < 0) {
    perror("generic.data_ack:sock_and_bind"); return false;
  }
  // (4) ctl ack port
  ctl_ack_listen_port = arg_data_listen_port + CTL_ACK_OFFSET;
  ctl_ack_sock_fd = sock_and_bind(ctl_ack_listen_port);
  if(ctl_ack_sock_fd < 0) {
    perror("generic.ctl_ack:sock_and_bind"); return false;
  }
  // (5) state port
  state_listen_port = arg_data_listen_port + STATE_OFFSET;
  state_sock_fd = sock_and_bind(state_listen_port);
  if(state_sock_fd < 0) {
    perror("generic.state:sock_and_bind"); return false;
  }
  return true;
} // End generic_client_info::init_sock()
