/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef GENERIC_CLIENT_INFO_H
#define GENERIC_CLIENT_INFO_H

#include <vector>
#include <stdio.h>
#include <netinet/in.h>

#include "../util/Debug.h"
#include "../util/LogLine.h"
#include "../udp-lib/ReqRespBlock.h"
#include "../serv-comp/ServCompIFPkt.h"

/** Defines the client of a generic service */

#define DATA_OFFSET 0
#define CTL_OFFSET 1
#define DATA_ACK_OFFSET 2
#define CTL_ACK_OFFSET 3
#define STATE_OFFSET 4

enum cl_state_t {
  IDLE_CL=0, // am idle
  PKT_WAIT_CL=1, // not yet recd the first data to be processed
  PROCESSING_CL=2, // processing the data received
};

class generic_client_info {
public:
  static Debug_t *dbg; // for debugging
  static char name[128]; // same as the one in the service structure

  static struct in_addr myIP;
  static reqRespBlock_t *rrBlock; // to send requests upstream
  static unsigned int PATH_HB_PERIOD; // in ms
  static unsigned int PATH_HB_TIMEOUT; // in ms

  int pathID; // the pathID of the path
  unsigned short path_dest_scid; // the dest scid of the path
  unsigned short version; // path version

  struct in_addr destIP; // where i'm supposed to send pkts
  int destPort; // where i'm supposed to send pkts

  int data_listen_port; // this is where i listen for RDM pkts
  int data_sock_fd; // this is the socket for listening

  int data_ack_listen_port; // this is where i listen for data acks
  int data_ack_sock_fd; // for acks on data_ack_listen_port

  int ctl_listen_port; // this is the control port (RDM)
  int ctl_sock_fd; // this is the socket for incoming ctl pkts

  int ctl_ack_listen_port; // this is the control ack port (RDM)
  int ctl_ack_sock_fd; // for acks on ctl_ack_listen_port

  int state_listen_port; // to listen for PATH_STATE_F
  int state_sock_fd; // to listen for PATH_STATE_F

  struct in_addr upstreamIP; // where i'm supposed to get data from
  int upstreamPort; // where i'm supposed to get data from

  unsigned short serv_dest_scid; // the destn SCID
  unsigned short serv_my_scid; // my SCID
  unsigned short serv_upstream_scid; // the upstream SCID

  UDPLibPkt_t reqPkt, ackPkt; // for communication with upstream
  int reqNo; // for communication with upstream

  UDPLibPkt_t ctl_reqPkt, ctl_ackPkt; // for sending soft-state
  int ctl_reqNo; // to send soft-state updates downstream

  int numHBSent, numHBRecd;
  struct timeval last_hb_send_time, last_hb_recv_time;

  enum cl_state_t state; // state of this client

  generic_client_info() { init_fields(); }

  /** Handle incoming data pkt */
  virtual void handleData() {}
  /** Handle incoming ctl pkt (APP_UPSTREAM_INFO/APP_HB/APP_KILL) */
  virtual void handleCtl();
  /** This is actually ack for the request for TEXT sent upstream */
  virtual void handleDataAck() {}
  /** This is the ack for the upstream info sent downstream */
  virtual void handleCtlAck();
  /** This is PATH_STATE_F being passed downstream */
  virtual void handleSoftState() {}

  /** Check for data sends */
  virtual void checkDataSend() {}
  /** Check for ctl sends */
  virtual void checkCtlSend();
  /** Check for soft-state sends */
  virtual void checkSoftStateSend() {}
  /** Check for heart-beat sends */
  virtual void checkHBSend();
  /** Check for timeout of path */
  virtual void checkPathTimeout();

  /** Send upstream info for the path elt downstream */
  virtual void sendUpstreamInfo();
  /** Send soft-state info downstream */
  virtual void sendSoftState() {}
  /** Send APP_HB/APP_KILL upstream */
  virtual void sendHB(ServCompPktType_t type);

  /** Discard the info on this -- this client is no more valid */
  virtual void endSession();

  /** Some utility routines */
  virtual bool parseDestSpec(const char *dest_ip_port);
  virtual bool parseIPPort(const char *dest_ip_port,
			   struct in_addr *ip, int *port);
  virtual void parseUpstreamInfo(ServCompIFPkt_t *sPkt,
				 ServCompIFPkt_t *resp_sPkt);

  /** Init local variables using the SC create pkt recd */
  virtual void init_from_sc(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt,
			    int arg_data_listen_port);
  virtual bool init_sock(int arg_data_listen_port);
  /** Init to default values */
  virtual void init_fields();

  virtual ~generic_client_info() {}
}; // End class generic_client_info
typedef class generic_client_info generic_client_info_t;

#endif
