/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef GENERIC_SERVICE_H
#define GENERIC_SERVICE_H

/** Defines a generic service, from which other services can be
    'dervived' */

#include "generic_client_info.h"

class generic_service {
protected:
  Debug_t *dbg; // for debugging/logging

  struct in_addr myIP; // to return to SC
  int udp_sock; // for pkts from SC
  generic_client_info_t *cl_list; // the list of clients
  int max_cl; // the max number of clients in the array structure

  int sc_listen_port; // the port on which to listen for SC pkts
  int data_port_base; // the port on which to listen for app-level pkts

  virtual generic_client_info_t &get_cl(int cl_index) {
    return cl_list[cl_index];
  }

  virtual JString myName();
  virtual void usage(const char *progname);
  /** Utility routine */
  virtual int findCl(ServCompIFPkt_t *sPkt);
  virtual int getFreeCl();

  /** Handle a msg from SC */
  virtual void handleSC();
  /** Handle create msg from SC */
  virtual void handleCreate(ServCompIFPkt_t *sPkt,
			    ServCompIFPkt_t *resp_sPkt) {
    handleCreate(sPkt, resp_sPkt, PKT_WAIT_CL);
  }
  /** Handle create msg from SC, and change state to given state */
  virtual void handleCreate(ServCompIFPkt_t *sPkt,
			    ServCompIFPkt_t *resp_sPkt,
			    enum cl_state_t next_state);
  /** Returns true if NO further processing is to be done, false otherwise */
  virtual bool checkDup(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt);
  /** Handle change-service-instance cmd from SC */
  virtual void handleCSI(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt);
  /** Handle kill cmd from SC */
  virtual void handleKill(ServCompIFPkt_t *sPkt, ServCompIFPkt_t *resp_sPkt);

  /** This is the interrupt handler */
  static void intr_catcher(int);
  static bool should_listen;
public:
  generic_service(generic_client_info_t *arg_cl_list, int arg_max_cl,
		  Debug_t *arg_dbg);
  /** To be called from outside */
  virtual int main(int argc, char *argv[]);
  virtual ~generic_service() {}
}; // End class generic_service
typedef class generic_service generic_service_t;

#endif
