/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements email_client_info.h */

#include "email_client_info.h"

#include "../udp-lib/UDPLib.h"
#include "../util/Util.h"

vector<JString> email_client_info::news_str_vec;

void email_client_info::handleData() {
  // recv the pkt
  UDPLibPkt_t uPkt; ServCompIFPkt_t sPkt;
  int recv_len; struct sockaddr_in from_addr;
  unsigned long rcv_count = 0; bool pkt_ready = true;
  int res = UDPLib::blockingRecv(data_sock_fd, &uPkt, &recv_len,
				 &from_addr, NULL, &rcv_count, pkt_ready);
  if(res != 1) { cerr << "res = " << res << endl; return; }
  // parse it
  int parseRes = sPkt.parsePkt(uPkt.data, recv_len);
  if(parseRes != 1) { cerr << "parseRes = " << parseRes << endl; return; }

  const char *start_str = sPkt.getValue(REQ_ARGS_F);
  if(start_str == NULL) {
    cerr << "no start_str found in request\n";
    exit(99);
  }
  int sentence_index = atoi(start_str);
  int numSentences = news_str_vec.size();
  if(sentence_index >= numSentences) {
    cerr << "warning: sentence_index=" << sentence_index
	 << " numSentences=" << numSentences << endl;
    sentence_index = numSentences-1;
  }

  ServCompIFPkt_t resp_sPkt;
  if(sPkt.fixed.type == TTS_REQ) {
    resp_sPkt.fixed.type = TTS_RESP;
    resp_sPkt.fixed.pathID = pathID;
    resp_sPkt.fixed.path_dest_scid = path_dest_scid;
    resp_sPkt.fixed.version = version;
    resp_sPkt.addPair(RESP_TEXT_F, news_str_vec[sentence_index].cstr());
    resp_sPkt.fixed.success = 1;
    dbg->print() << name << " tts-req p_id=" << pathID << ':'
		 << path_dest_scid << " path_ver=" << version << endl;
  } else {
    cerr << "email_cl.handleData: unknown type ["
	 << ServCompIFPkt::type2str((ServCompPktType_t)sPkt.fixed.type)
	 << "]\n";
    resp_sPkt.fixed.success = 0;
  }

  // send response
  int send_len = resp_sPkt.toStr(uPkt.data, sizeof(uPkt.data));
  int send_fd = data_ack_sock_fd; // can use same sock to send pkts too
  // should send pkt via the delaying tunnel
  bool use_tunnel = (serv_my_scid != serv_dest_scid);
  unsigned long count = 0;
  res = UDPLib::sendResp(send_fd, &uPkt, send_len,
			 from_addr.sin_addr, destPort+DATA_ACK_OFFSET,
			 use_tunnel, &count);
  if(res != 1) {
    cerr << "email_cl.handleData: res = " << res << endl; return;
  }
} // End email_client_info::handleData()
