/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements Peer.h */

#include "Peer.h"

bool Peer::use_queue = false;

queue<int, list<int> > Peer::send_reqNoQueue;
vector<bool> Peer::send_reqNo_vec;
unsigned int Peer::send_lastRet = 0;
vector<int> Peer::send_vec;

queue<int, list<int> > Peer::fwd_reqNoQueue;
vector<bool> Peer::fwd_reqNo_vec;
unsigned int Peer::fwd_lastRet = 0;
vector<int> Peer::fwd_vec;

queue<ackInfo_t, list<ackInfo_t> > Peer::sc_reqNoQueue;
vector<active_ackInfo_t> Peer::sc_reqNo_vec;
unsigned int Peer::sc_lastRet = 0;
vector<ackInfo_t> Peer::sc_vec;

queue<repair_ackInfo_t, list<repair_ackInfo_t> > Peer::sc_repair_reqNoQueue;
vector<active_repair_ackInfo_t> Peer::sc_repair_reqNo_vec;
unsigned int Peer::sc_repair_lastRet = 0;
vector<repair_ackInfo_t> Peer::sc_repair_vec;

queue<ackInfo_t, list<ackInfo_t> > Peer::sc_alt_reqNoQueue;
vector<active_ackInfo_t> Peer::sc_alt_reqNo_vec;
unsigned int Peer::sc_alt_lastRet = 0;
vector<ackInfo_t> Peer::sc_alt_vec;

queue<ackInfo_t, list<ackInfo_t> > Peer::sc_serv_reqNoQueue;
vector<active_ackInfo_t> Peer::sc_serv_reqNo_vec;
unsigned int Peer::sc_serv_lastRet = 0;
vector<ackInfo_t> Peer::sc_serv_vec;

queue<int, list<int> > Peer::sc_hb_reqNoQueue;
vector<bool> Peer::sc_hb_reqNo_vec;
unsigned int Peer::sc_hb_lastRet = 0;
vector<int> Peer::sc_hb_vec;

void Peer::addReq(int reqNo, queue<int, list<int> > *reqNoQueue,
		  vector<bool> *reqNo_vec) {
  if(use_queue) {
    reqNoQueue->push(reqNo);
  } else {
    for(;;) {
      int curr_size = reqNo_vec->size();
      if(reqNo < curr_size) { break; }
      bool flag = false; reqNo_vec->push_back(flag);
    }
    (*reqNo_vec)[reqNo] = true;
  }
} // End Peer:addReq

int Peer::popReqNos(int *reqNos, int maxPop,
		    queue<int, list<int> > *reqNoQueue,
		    vector<bool> *reqNo_vec, unsigned int *lastRet) {
  int numPopped = 0;
  if(use_queue) {
    for(int i = 0; i < maxPop; i++) {
      if(reqNoQueue->empty()) break;
      reqNos[numPopped] = reqNoQueue->front();
      reqNoQueue->pop();
      numPopped++;
    }

    // end case when use_queue is true
  } else {

    for(unsigned int i = 0; i < reqNo_vec->size(); i++) {
      if(numPopped == maxPop) { break; }
      if((*reqNo_vec)[*lastRet]) {
	reqNos[numPopped] = *lastRet;
	numPopped++;
	(*reqNo_vec)[*lastRet] = false;
      }
      (*lastRet)++;
      if(*lastRet == reqNo_vec->size()) {
	*lastRet = 0;
      }
    }

    // end case when use_queue is false
  }
  return numPopped;
} // End Peer::popReqNos()

int Peer::vec_pop(queue<int, list<int> > *reqNoQueue,
		  vector<bool> *reqNo_vec, unsigned int *lastRet,
		  vector<int> *vec) {
  unsigned int numPopped = 0;
  if(use_queue) {
    while(!reqNoQueue->empty()) {
      if(numPopped == vec->size()) {
	int new_reqNo = -1; vec->push_back(new_reqNo);
      }
      (*vec)[numPopped] = reqNoQueue->front();
      reqNoQueue->pop();
      numPopped++;
    }

    // end case when use_queue is true
  } else {
    for(unsigned int i = 0; i < reqNo_vec->size(); i++) {
      if((*reqNo_vec)[*lastRet]) {
	if(numPopped == vec->size()) {
	  int new_reqNo = -1; vec->push_back(new_reqNo);
	}
	(*vec)[numPopped] = *lastRet;
	numPopped++;
	(*reqNo_vec)[*lastRet] = false;
      }
      (*lastRet)++;
      if(*lastRet == reqNo_vec->size()) {
	*lastRet = 0;
      }
    }

    // end case when use_queue is false
  }
  return numPopped;
} // End Peer::vec_pop()

void Peer::a_addReq(ackInfo_t &aInfo,
		    queue<ackInfo_t, list<ackInfo_t> > *reqNoQueue,
		    vector<active_ackInfo_t> *reqNo_vec) {
  if(use_queue) {
    reqNoQueue->push(aInfo);
  } else {
    // increase size of sc_serv_reqNo_vec to what is necessary
    int reqNo = aInfo.reqNo;
    for(;;) {
      int curr_size = reqNo_vec->size();
      if(reqNo < curr_size) { break; }
      active_ackInfo_t a_aInfo; reqNo_vec->push_back(a_aInfo);
    }
    (*reqNo_vec)[reqNo].active = true;
    (*reqNo_vec)[reqNo].aInfo = aInfo;
  }
} // End Peer::a_addReq()

int Peer::a_popReqNos(ackInfo_t *aVec, int maxPop,
		      queue<ackInfo_t, list<ackInfo_t> > *reqNoQueue,
		      vector<active_ackInfo_t> *reqNo_vec,
		      unsigned int *lastRet) {
  int numPopped = 0;
  if(use_queue) {
    for(int i = 0; i < maxPop; i++) {
      if(reqNoQueue->empty()) break;
      aVec[numPopped] = reqNoQueue->front();
      reqNoQueue->pop();
      numPopped++;
    }
    // end case when use_queue is true
  } else {

    for(unsigned int i = 0; i < reqNo_vec->size(); i++) {
      if(numPopped == maxPop) { break; }
      active_ackInfo_t &a_aInfo = (*reqNo_vec)[*lastRet];
      if(a_aInfo.active) {
	aVec[numPopped] = a_aInfo.aInfo;
	numPopped++;
	a_aInfo.active = false;
      }
      (*lastRet)++;
      if(*lastRet == reqNo_vec->size()) {
	*lastRet = 0;
      }
    }

    // end case when use_queue is false
  }
  return numPopped;
} // End Peer::a_popReqNos()

int Peer::a_vec_pop(queue<ackInfo_t, list<ackInfo_t> > *reqNoQueue,
		    vector<active_ackInfo_t> *reqNo_vec,
		    unsigned int *lastRet, vector<ackInfo_t> *vec) {
  unsigned int numPopped = 0;
  if(use_queue) {
    while(!reqNoQueue->empty()) {
      if(numPopped == vec->size()) {
	ackInfo_t aInfo; vec->push_back(aInfo);
      }
      (*vec)[numPopped] = reqNoQueue->front();
      reqNoQueue->pop();
      numPopped++;
    }

    // end case when use_queue is true
  } else {
    for(unsigned int i = 0; i < reqNo_vec->size(); i++) {
      active_ackInfo_t &a_aInfo = (*reqNo_vec)[*lastRet];
      if(a_aInfo.active) {
	if(numPopped == vec->size()) {
	  ackInfo_t aInfo; vec->push_back(aInfo);
	}
	(*vec)[numPopped] = a_aInfo.aInfo;
	numPopped++;
	a_aInfo.active = false;
      }
      (*lastRet)++;
      if(*lastRet == reqNo_vec->size()) {
	*lastRet = 0;
      }
    }

    // end case when use_queue is false
  }
  return numPopped;
} // End Peer::a_vec_pop()

void Peer::a_rep_addReq(repair_ackInfo_t &aInfo,
			queue<repair_ackInfo_t, list<repair_ackInfo_t> > *reqNoQueue,
			vector<active_repair_ackInfo_t> *reqNo_vec) {
  if(use_queue) {
    reqNoQueue->push(aInfo);
  } else {
    // increase size of sc_serv_reqNo_vec to what is necessary
    int reqNo = aInfo.reqNo;
    for(;;) {
      int curr_size = reqNo_vec->size();
      if(reqNo < curr_size) { break; }
      active_repair_ackInfo_t a_aInfo; reqNo_vec->push_back(a_aInfo);
    }
    (*reqNo_vec)[reqNo].active = true;
    (*reqNo_vec)[reqNo].rep_aInfo = aInfo;
  }
} // End Peer::a_rep_addReq()

int Peer::a_rep_popReqNos(repair_ackInfo_t *aVec, int maxPop,
			  queue<repair_ackInfo_t, list<repair_ackInfo_t> > *reqNoQueue,
			  vector<active_repair_ackInfo_t> *reqNo_vec,
			  unsigned int *lastRet) {
  int numPopped = 0;
  if(use_queue) {
    for(int i = 0; i < maxPop; i++) {
      if(reqNoQueue->empty()) break;
      aVec[numPopped] = reqNoQueue->front();
      reqNoQueue->pop();
      numPopped++;
    }
    // end case when use_queue is true
  } else {

    for(unsigned int i = 0; i < reqNo_vec->size(); i++) {
      if(numPopped == maxPop) { break; }
      active_repair_ackInfo_t &a_aInfo = (*reqNo_vec)[*lastRet];
      if(a_aInfo.active) {
	aVec[numPopped] = a_aInfo.rep_aInfo;
	numPopped++;
	a_aInfo.active = false;
      }
      (*lastRet)++;
      if(*lastRet == reqNo_vec->size()) {
	*lastRet = 0;
      }
    }

    // end case when use_queue is false
  }
  return numPopped;
} // End Peer::a_rep_popReqNos()

int Peer::a_rep_vec_pop(queue<repair_ackInfo_t, list<repair_ackInfo_t> > *reqNoQueue,
			vector<active_repair_ackInfo_t> *reqNo_vec,
			unsigned int *lastRet, vector<repair_ackInfo_t> *vec) {
  unsigned int numPopped = 0;
  if(use_queue) {
    while(!reqNoQueue->empty()) {
      if(numPopped == vec->size()) {
	repair_ackInfo_t aInfo; vec->push_back(aInfo);
      }
      (*vec)[numPopped] = reqNoQueue->front();
      reqNoQueue->pop();
      numPopped++;
    }

    // end case when use_queue is true
  } else {
    for(unsigned int i = 0; i < reqNo_vec->size(); i++) {
      active_repair_ackInfo_t &a_aInfo = (*reqNo_vec)[*lastRet];
      if(a_aInfo.active) {
	if(numPopped == vec->size()) {
	  repair_ackInfo_t aInfo; vec->push_back(aInfo);
	}
	(*vec)[numPopped] = a_aInfo.rep_aInfo;
	numPopped++;
	a_aInfo.active = false;
      }
      (*lastRet)++;
      if(*lastRet == reqNo_vec->size()) {
	*lastRet = 0;
      }
    }

    // end case when use_queue is false
  }
  return numPopped;
} // End Peer::a_rep_vec_pop()

void Peer::send_addReq(int reqNo) {
  addReq(reqNo, &send_reqNoQueue, &send_reqNo_vec);
} // End Peer::send_addReq()

int Peer::send_popReqNos(int *reqNos, int maxPop) {
  return popReqNos(reqNos, maxPop, &send_reqNoQueue, &send_reqNo_vec,
		   &send_lastRet);
} // End Peer::send_popReqNos()

int Peer::send_vec_pop() {
  return vec_pop(&send_reqNoQueue, &send_reqNo_vec, &send_lastRet,
		 &send_vec);
} // End Peer::send_vec_pop()

void Peer::fwd_addReq(int reqNo) {
  addReq(reqNo, &fwd_reqNoQueue, &fwd_reqNo_vec);
} // End Peer::fwd_addReq()

int Peer::fwd_popReqNos(int *reqNos, int maxPop) {
  return popReqNos(reqNos, maxPop, &fwd_reqNoQueue, &fwd_reqNo_vec,
		   &fwd_lastRet);
} // End Peer::fwd_popReqNos()

int Peer::fwd_vec_pop() {
  return vec_pop(&fwd_reqNoQueue, &fwd_reqNo_vec, &fwd_lastRet,
		 &fwd_vec);
} // End Peer::fwd_vec_pop()

void Peer::sc_hb_addReq(int reqNo) {
  addReq(reqNo, &sc_hb_reqNoQueue, &sc_hb_reqNo_vec);
} // End Peer::sc_hb_addReq()

int Peer::sc_hb_popReqNos(int *reqNos, int maxPop) {
  return popReqNos(reqNos, maxPop, &sc_hb_reqNoQueue,
		   &sc_hb_reqNo_vec, &sc_hb_lastRet);
} // End Peer::sc_hb_popReqNos()

int Peer::sc_hb_vec_pop() {
  return vec_pop(&sc_hb_reqNoQueue, &sc_hb_reqNo_vec, &sc_hb_lastRet,
		 &sc_hb_vec);
} // End Peer::sc_hb_vec_pop()

void Peer::send_addServInfoReq(int reqNo) {
  addReq(reqNo, &send_reqNoQueueServInfo, &send_reqNoServInfo_vec);
} // End Peer::send_addServInfoReq()

int Peer::send_popReqNosServInfo(int *reqNos, int maxPop) {
  return popReqNos(reqNos, maxPop, &send_reqNoQueueServInfo,
		   &send_reqNoServInfo_vec, &send_ServInfo_lastRet);
} // End Peer::send_popReqNosServInfo()

void Peer::fwd_addServInfoReq(int reqNo) {
  addReq(reqNo, &fwd_reqNoQueueServInfo, &fwd_reqNoServInfo_vec);
} // End Peer::fwd_addServInfoReq()

int Peer::fwd_popReqNosServInfo(int *reqNos, int maxPop) {
  return popReqNos(reqNos, maxPop, &fwd_reqNoQueueServInfo,
		   &fwd_reqNoServInfo_vec, &fwd_ServInfo_lastRet);
} // End Peer::fwd_popReqNosServInfo()

void Peer::sc_addReq(int reqNo, UDPLibPkt_t *uPkt, struct in_addr toReplyAddr,
		     int toReplyPort, int pathID, unsigned int path_dest_scid,
		     bool to_app = false) {
  ackInfo_t aInfo;
  aInfo.reqNo = reqNo; 		aInfo.uPkt = uPkt;
  aInfo.toReplyAddr = toReplyAddr; 	aInfo.toReplyPort = toReplyPort;
  aInfo.pathID = pathID;		aInfo.path_dest_scid = path_dest_scid;
  aInfo.to_app = to_app;
  a_addReq(aInfo, &sc_reqNoQueue, &sc_reqNo_vec);
} // End Peer::sc_addReq()

int Peer::sc_popReqNos(ackInfo_t *aVec, int maxPop) {
  return a_popReqNos(aVec, maxPop, &sc_reqNoQueue, &sc_reqNo_vec,
		     &sc_lastRet);
} // End Peer::sc_popReqNos()

int Peer::sc_vec_pop() {
  return a_vec_pop(&sc_reqNoQueue, &sc_reqNo_vec, &sc_lastRet, &sc_vec);
} // End Peer::sc_vec_pop()

void Peer::sc_repair_addReq(int reqNo, UDPLibPkt_t *uPkt,
			    struct in_addr toReplyAddr,
			    int toReplyPort, int pathID,
			    unsigned int path_dest_scid,
			    unsigned short fail_dest_scid,
			    unsigned short fail_origin_scid,
			    short posn, bool local) {
  repair_ackInfo_t aInfo;
  aInfo.reqNo = reqNo; 			aInfo.uPkt = uPkt;
  aInfo.toReplyAddr = toReplyAddr; 	aInfo.toReplyPort = toReplyPort;
  aInfo.pathID = pathID;		aInfo.path_dest_scid = path_dest_scid;
  aInfo.fail_dest_scid = fail_dest_scid;
  aInfo.fail_origin_scid = fail_origin_scid;
  aInfo.posn = posn;			aInfo.local = local;
  a_rep_addReq(aInfo, &sc_repair_reqNoQueue, &sc_repair_reqNo_vec);
} // End Peer::sc_repair_addReq()

int Peer::sc_repair_popReqNos(repair_ackInfo_t *aVec, int maxPop) {
  return a_rep_popReqNos(aVec, maxPop, &sc_repair_reqNoQueue,
			 &sc_repair_reqNo_vec, &sc_repair_lastRet);
} // End Peer::sc_repair_popReqNos()

int Peer::sc_repair_vec_pop() {
  return a_rep_vec_pop(&sc_repair_reqNoQueue, &sc_repair_reqNo_vec,
		       &sc_repair_lastRet, &sc_repair_vec);
} // End Peer::sc_repair_vec_pop()

void Peer::sc_alt_addReq(int reqNo, UDPLibPkt_t *uPkt,
			 struct in_addr toReplyAddr, int toReplyPort,
			 int pathID, unsigned int path_dest_scid) {
  ackInfo_t aInfo;
  aInfo.reqNo = reqNo; 			aInfo.uPkt = uPkt;
  aInfo.toReplyAddr = toReplyAddr; 	aInfo.toReplyPort = toReplyPort;
  aInfo.pathID = pathID;		aInfo.path_dest_scid = path_dest_scid;
  a_addReq(aInfo, &sc_alt_reqNoQueue, &sc_alt_reqNo_vec);
} // End Peer::sc_alt_addReq()

int Peer::sc_alt_popReqNos(ackInfo_t *aVec, int maxPop) {
  return a_popReqNos(aVec, maxPop, &sc_alt_reqNoQueue,
		     &sc_alt_reqNo_vec, &sc_alt_lastRet);
} // End Peer::sc_alt_popReqNos()

int Peer::sc_alt_vec_pop() {
  return a_vec_pop(&sc_alt_reqNoQueue, &sc_alt_reqNo_vec,
		   &sc_alt_lastRet, &sc_alt_vec);
} // End Peer::sc_alt_vec_pop()

void Peer::sc_serv_addReq(int reqNo, UDPLibPkt_t *uPkt,
			  struct in_addr toReplyAddr, int toReplyPort,
			  int pathID, unsigned int path_dest_scid,
			  unsigned short version, bool to_app) {
  ackInfo_t aInfo;
  aInfo.reqNo = reqNo; 			aInfo.uPkt = uPkt;
  aInfo.toReplyAddr = toReplyAddr; 	aInfo.toReplyPort = toReplyPort;
  aInfo.pathID = pathID;		aInfo.path_dest_scid = path_dest_scid;
  aInfo.version = version;		aInfo.to_app = to_app;
  a_addReq(aInfo, &sc_serv_reqNoQueue, &sc_serv_reqNo_vec);
} // End Peer::sc_serv_addReq()

int Peer::sc_serv_popReqNos(ackInfo_t *aVec, int maxPop) {
  return a_popReqNos(aVec, maxPop, &sc_serv_reqNoQueue,
		     &sc_serv_reqNo_vec, &sc_serv_lastRet);
} // End Peer::sc_serv_popReqNos()

int Peer::sc_serv_vec_pop() {
  return a_vec_pop(&sc_serv_reqNoQueue, &sc_serv_reqNo_vec,
		   &sc_serv_lastRet, &sc_serv_vec);
} // End Peer::sc_serv_vec_pop()

void Peer::addPath(int pathID, unsigned short path_dest_scid) {
  path_id_t p_id = { pathID, path_dest_scid };
  for(unsigned int i = 0; i < path_list.size(); i++) {
    if((path_list[i].pathID == pathID) &&
       (path_list[i].path_dest_scid == path_dest_scid)) {
      return;
    }
  }
  path_list.push_back(p_id);
} // End Peer::addPath()

void Peer::delPath(int pathID, unsigned short path_dest_scid) {
  vector<path_id_t>::iterator iter;
  for(iter = path_list.begin(); iter != path_list.end(); iter++) {
    if((iter->pathID == pathID) &&
       (iter->path_dest_scid == path_dest_scid)) {
      path_list.erase(iter);
      break;
    }
  }
} // End Peer::delPath()
