/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements GenLat.h */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

#include "GenLat.h"
#include "../util/Util.h"

unsigned long GenLat::small_runLength = 60*1000;

GenLat::GenLat(unsigned short arg_baseline, unsigned long arg_runLength,
	       Debug_t *arg_dbg) {
  dbg = arg_dbg;

  if(arg_runLength < small_runLength) {
    fprintf(stderr, "Warning: arg_runLength (%ld) < small_runLength (%ld)\n",
	    arg_runLength, small_runLength);
  }

  baseline = arg_baseline;
  runLength = arg_runLength;

  // set random seed
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long seedLong = tv.tv_usec%211;
  float seedFloat = ((seedLong+1)*1.0)/212;
  fprintf(stderr, "setting random seed done with %ld(%f)\n",
	  seedLong, seedFloat);
  Uniform::Set(seedFloat);

  // set the current and next average -- somewhere between baseline
  // and 2*baseline
  currAvg = baseline + (unsigned short)(U.Next()*baseline);
  nextAvg = baseline + (unsigned short)(U.Next()*baseline);
  gettimeofday(&currAvgStartTime, NULL);

  // set the current 1min avg
  long val = currAvg + (long)(U.Next()*21) - 10;
  if(val > 0) { curr_1min_avg = val; }
  else { curr_1min_avg = currAvg; }
  gettimeofday(&curr_1min_avg_start_time, NULL);

} // End GenLat::GenLat()

unsigned short GenLat::currLat() {
  float gap = now(currAvgStartTime);
  float gap1 = now(curr_1min_avg_start_time);

  if(gap > 2*runLength) {
    // set the current and next average -- somewhere between baseline
    // and 2*baseline
    currAvg = baseline + (unsigned short)(U.Next()*baseline);
    nextAvg = baseline + (unsigned short)(U.Next()*baseline);
    gettimeofday(&currAvgStartTime, NULL);

    // set the current 1min avg
    long val = currAvg + (long)(U.Next()*21) - 10;
    if(val > 0) { curr_1min_avg = val; }
    else { curr_1min_avg = currAvg; }
    gettimeofday(&curr_1min_avg_start_time, NULL);

    //fprintf(stderr, "*1 currAvg=%d, nextAvg=%d, curr_1min_avg=%d\n",
    //currAvg, nextAvg, curr_1min_avg);

    // return something within 10ms in either direction of
    // curr_1min_avg
    long ret_val = curr_1min_avg + (long)(U.Next()*21) - 10;
    if(ret_val > 0) { return ret_val; }
    else { return curr_1min_avg; }

  } else if(gap > runLength) {
    // set the current and next average
    currAvg = nextAvg;
    nextAvg = baseline + (unsigned short)(U.Next()*baseline);
    gettimeofday(&currAvgStartTime, NULL);

    // set the current 1min avg
    long val = currAvg + (long)(U.Next()*21) - 10;
    if(val > 0) { curr_1min_avg = val; }
    else { curr_1min_avg = currAvg; }
    gettimeofday(&curr_1min_avg_start_time, NULL);

    //fprintf(stderr, "*2 currAvg=%d, nextAvg=%d, curr_1min_avg=%d\n",
    //currAvg, nextAvg, curr_1min_avg);

    // return something within 10ms in either direction of
    // curr_1min_avg
    long ret_val = curr_1min_avg + (long)(U.Next()*21) - 10;
    if(ret_val > 0) { return ret_val; }
    else { return curr_1min_avg; }

  } else if(gap1 > small_runLength) {
    long num_steps_avail =
      (runLength - (long)now(currAvgStartTime))/(small_runLength);
    long next_val;
    if(num_steps_avail != 0) {
      next_val = curr_1min_avg + (nextAvg-curr_1min_avg)/num_steps_avail;
    } else {
      next_val = nextAvg;
    }
    if(next_val > 0) { curr_1min_avg = next_val; }
    else { curr_1min_avg = currAvg; }
    gettimeofday(&curr_1min_avg_start_time, NULL);

    //fprintf(stderr, "*3 currAvg=%d, nextAvg=%d, curr_1min_avg=%d\n",
    //currAvg, nextAvg, curr_1min_avg);

    // return something within 10ms in either direction of
    // curr_1min_avg
    long val = curr_1min_avg + (long)(U.Next()*21) - 10;
    if(val > 0) { return val; }
    else { return curr_1min_avg; }

  } else {
    // return something within 10ms in either direction of
    // curr_1min_avg
    long val = curr_1min_avg + (long)(U.Next()*21) - 10;
    if(val > 0) { return val; }
    else { return curr_1min_avg; }

  }

} // End GenLat::currLat()
