/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef GENLAT_H
#define GENLAT_H

#include <sys/time.h>

#include "../util/Debug.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"

/** Class to generate latencies between two hosts separated over the
    wide-area */
class GenLat {
protected:
  /** The debug structure */
  Debug_t *dbg;

  /** The baseline latency, in ms */
  unsigned short baseline;
  /** The runLength of the "low variation" period, in ms */
  unsigned long runLength;

  /** The current 1 minute average */
  unsigned short curr_1min_avg;
  /** The start time of 'curr_1min_avg' */
  struct timeval curr_1min_avg_start_time;

  /** The average in this run-length, of the 'curr_1min_avg' values,
      in ms */
  unsigned short currAvg;
  /** The average latency for the next run-length */
  unsigned short nextAvg;

  /** The time when the current average latency (over 1min intervals)
      started */
  struct timeval currAvgStartTime;

  /** The small-length 1min interval */
  static unsigned long small_runLength;

  /** From the newran02 library, to generate uniform random numbers */
  Uniform U;

public:
  /** Init the object, given the baseline latency, and the runLength
      of the "low variation" period.  Both are given in ms. */
  GenLat(unsigned short arg_baseline, unsigned long arg_runLength,
	 Debug_t *arg_dbg);

  /** Return the current latency, in ms */
  unsigned short currLat();

}; // End class GenLat

typedef class GenLat GenLat_t;

#endif
