/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef PKTDROP_H
#define PKTDROP_H

#include <sys/time.h>

#include "../util/Debug.h"
#include "../util/JString.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"

typedef struct Entry {
  unsigned long t; // the gap value, in ms
  float F; // F(t), where F is the CDF
} Entry_t;

/** This class implements the state machine to determine packet drops
    according to a given distribution of pkt loss patterns */
class PktDrop {
protected:
  /** The debug structure */
  Debug_t *dbg;
  /** The scids pair for which this PktDrop is used */
  unsigned short dest_scid, origin_scid;

  /** Stores the (discrete) CDF */
  Entry_t *CDF;
  /** The number of entries in the CDF */
  int numEntries;

  /** Stores the current gap value, in ms */
  unsigned long currGap;
  /** The start time of the current gap */
  struct timeval currGapStart;

  /** Gives the inverse of the CDF, defined as:
      CDF_inv(y) = inf { x: F(x) >= y }. */
  unsigned long CDF_inv(float y);

  /** From newran02 library, to generate uniform random numbers */
  Uniform U;

public:
  /** The distribution file is given as argument to the constructor.
      The file has the probability density function in the following
      format: in each line: "gap-in-ms frequency probability".  This
      means that gaps of the given length happen with the given
      probability.  (The frequency is from the particular experiment,
      and can be ignored). */
  PktDrop(const char *distFileName, unsigned short a_dest_scid,
	  unsigned short a_origin_scid, Debug_t *arg_dbg);

  /** The destructor */
  ~PktDrop();

  /** Determine whether or not a packet should be dropped now, based
      on the distribution of loss-period frequencies. */
  bool dropNow();

  void print();

}; // End class PktDrop

typedef class PktDrop PktDrop_t;

#endif
