#!/usr/bin/perl

# (c) 2002 by Bhaskaran Raman and Regents of the University of California

# Program to calculate the difference between the latency supposedly
# introduced by 'fwdRaw' and the one measured by 'measFwdRaw'

sub usage {
    print STDERR "Usage: $0 <meas-file> <fwd-file>\n";
} # End usage()

MAIN: {
    ($#ARGV == 1) || do { usage(); exit(1); };
    my $measFile = shift;
    my $fwdFile = shift;

    print STDERR "measFile \t= $measFile\n";
    print STDERR "fwdFile \t= $fwdFile\n";

    open(MEAS_FILE, "<$measFile") || die "open $measFile: $!";
    open(FWD_FILE, "<$fwdFile") || die "open $fwdFile: $!";

    my $numMeas = 0; # The number of latencies measured
    # The sum of differences in latency
    my $sum = 0;
    # The sum of squares of differences in latency
    my $sumsq = 0;
    # The number of pkts dropped by fwdRaw
    my $numDropped = 0;
    # The number of pkts lost in addition to that dropped by fwdRaw
    my $numMissed = 0;
    my $numLines = 0; # The number of lines in each file (has to be equal)

    # The maximum difference in latency
    my $max = 0;
    # The number of cases with over 10ms addnl delay
    my $numOver10ms = 0;

  READLOOP:
    for(;;) {
	# read the lines from the two files
	my $measLine = <MEAS_FILE>;
	my $fwdLine = <FWD_FILE>;
	if(!defined($measLine) || !defined($fwdLine)) {
	    if(defined($measLine) || defined($fwdLine)) {
		print STDERR "Error: $measFile and $fwdFile do not".
		    " have the same number of lines\n";
		exit(2);
	    }
	    last READLOOP;
	}

	# get the fields
	my ($m_measIndex, $m_seqNo, $m_recd, $m_delay) =
	    split(/\s+/, $measLine);
	my ($f_measIndex, $f_seqNo, $f_recd, $f_delay) =
	    split(/\s+/, $fwdLine);

	# check fields
	if(($m_measIndex != $f_measIndex) ||
	   ($m_measIndex != $numLines)) {
	    print STDERR "Error: m_measIndex=$m_measIndex,".
		" f_measIndex=$f_measIndex, lineNum=$numLines\n";
	    exit(3);
	}
	if($m_seqNo != $f_seqNo) {
	    if($f_seqNo != 0) {
		print STDERR "Error: m_seqNo=$m_seqNo, f_seqNo=$f_seqNo".
		    ", lineNum=$numLines\n";
		exit(4);
	    }
	}
	if(($m_recd == 1) && ($f_recd == 0)) {
	    print STDERR "Error: Dropped packet recd?  lineNum=$numLines\n";
	    exit(5);
	}

	# update statistics
	if($m_recd == 1) {
	    # this is a proper measurement
	    my $diff = $m_delay - $f_delay;
	    if($diff < 0) {
		if($diff != -1) {
		    print STDERR "Warning: m_delay=$m_delay".
			", f_delay=$f_delay, lineNum=$numLines\n";
		}
		$diff = 0;
	    }
	    $sum += $diff;
	    $sumsq += $diff*$diff;
	    $numMeas++;
	    if($max < $diff) { $max = $diff; }
	    if($diff > 10) { $numOver10ms++; }
	} else {
	    if($f_recd == 1) {
		# this is the case when the pkt forwarded by fwdRaw
		# was lost
		$numMissed++;
	    } else {
		if($m_seqNo == $f_seqNo) {
		    # this is the case when fwdRaw decided to drop the pkt
		    $numDropped++;
		} else {
		    # this is the case when fwdRaw never got the pkt
		    $numMissed++;
		}
	    }
	}

	# go on to read the next line
	$numLines++;

    } # End for(;;) reading all lines

    # print statistics
    my $percMissed = (100.0*$numMissed)/$numLines;
    my $percDropped = (100.0*$numDropped)/$numLines;
    print STDERR "\n";
    print STDERR "number of probe pkts    = $numLines\n";
    print STDERR "number of dropped pkts  = $numDropped ($percDropped\%)\n";
    print STDERR "number of missed pkts   = $numMissed ($percMissed\%)\n";
    print STDERR "\n";

    my $avg = (1.0*$sum)/$numMeas;
    my $sd = (1.0*$sumsq)/$numMeas - $avg*$avg;
    my $se = sqrt($sd);
    my $percOver10ms = (100.0*$numOver10ms)/$numLines;
    print STDERR "avg addnl delay         = $avg\n";
    print STDERR "se for addnl delay      = $se\n";
    print STDERR "max addnl delay         = $max\n";
    print STDERR "#pkts >10ms addnl delay = $numOver10ms ($percOver10ms\%)\n";

    print STDOUT "$percMissed\t$avg\t$se\t$max\t$percOver10ms\n";

} # End MAIN
