/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to test PktDrop.cpp */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include "PktDrop.h"
#include "../util/Util.h"

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <distbn-file>\n", progname);
} // End usage()

const int numGaps = 25;
unsigned long gapCount[numGaps];
int sleepTime = 100; // in ms

void intr_catcher(int) {
  unsigned long tot_count = 0;
  for(int i = 0; i < numGaps; i++) {
    tot_count += gapCount[i];
  }
  cout << endl
       << "----------------------------------------------------------------------"
       << endl;
  for(int i = 0; i < numGaps; i++) {
    cout << (i*sleepTime) << " \t" << gapCount[i] << " \t"
	 << ((1.0*gapCount[i])/tot_count) << endl;
  }
} // End intr_catcher

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 2) { usage(argv[0]); exit(1); }

  // Init the debug object
  struct timeval startTime;
  gettimeofday(&startTime, NULL);
  char logFile[128];
  unsigned short mySCID = 0;
  sprintf(logFile, "/tmp/fwd-log.%d", mySCID);
  Debug_t dbg(logFile, mySCID, startTime, cfg_to_file, cfg_use_stream);

  const char *distFileName = argv[1];
  PktDrop_t pd(distFileName, 0, 0, &dbg);

  for(int i = 0; i < numGaps; i++) {
    gapCount[i] = 0;
  }

  // Activate interrupt catcher
  signal(SIGINT, intr_catcher);
  //signal(SIGTERM, intr_catcher);

  int dropRun = 0;

  for(;;) {
    if(pd.dropNow()) {
      dropRun++;
      cerr << "#";
    } else {
      gapCount[dropRun]++;
      dropRun = 0;
      cerr << ".";
    }
    usleep(sleepTime*1000); // sleep for 10ms
  }

} // End main()
