/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements noop_client_info.h */

#include "noop_client_info.h"

#include "../udp-lib/UDPLib.h"
#include "../util/Util.h"

void noop_client_info::handleData() {
  fwdPkt(data_sock_fd, DATA_OFFSET);
  fprintf(stderr, "+");
} // End noop_client_info::handleData()

void noop_client_info::fwdPkt(int fd, int port_offset) {
  struct sockaddr_in from_addr;
  UDPLibPkt_t upkt;

  // receive packet (blocking)
  // NOTE: called only when something is ready to be read
  int recv_len = lib_recvfrom(fd, &upkt, sizeof(upkt), &from_addr);
  if(recv_len == -1) { perror("noop_cl.fwdPkt.recvfrom"); return; }

  unsigned short pkt_dest_scid = LogLine::INVALID_SCID;
  struct sockaddr_in to_addr;
  bzero((char *)&to_addr, sizeof(to_addr));
  to_addr.sin_family = PF_INET;
  // from upstream ==> send downstream, vice versa
  bool upstream_scid_match = true;
  bool dest_scid_match = true;
#ifdef IP_TUNNEL
  upstream_scid_match = (ntohs(upkt.my_scid) == serv_upstream_scid);
  dest_scid_match = (ntohs(upkt.my_scid) == serv_dest_scid);
#endif
  if((from_addr.sin_addr.s_addr == upstreamIP.s_addr) &&
     upstream_scid_match) {
    pkt_dest_scid = serv_dest_scid;
    to_addr.sin_addr = destIP;
    to_addr.sin_port = htons(destPort+port_offset);
  } else if((from_addr.sin_addr.s_addr == destIP.s_addr) &&
	    dest_scid_match) {
    pkt_dest_scid = serv_upstream_scid;
    to_addr.sin_addr = upstreamIP;
    to_addr.sin_port = htons(upstreamPort+port_offset);
  } else {
    cerr << "noop.cl: pkt from unknown IP\n";
    dbg->print() << name << " unknown-pkt p_id=" << pathID << ':'
		 << path_dest_scid << " path_ver=" << version
		 << " serv_my_scid=" << serv_my_scid
		 << " serv_dest_scid=" << serv_dest_scid
		 << " serv_upstream_scid=" << serv_upstream_scid
		 << " upkt_dest_scid=" << upkt.dest_scid
		 << " upkt_my_scid=" << upkt.my_scid << endl;
    return;
  }

  // forward data
  bool use_tunnel = ((pkt_dest_scid != serv_my_scid) &&
		     (pkt_dest_scid != LogLine::INVALID_SCID));
  unsigned long numSentDummy;
  int send_len = recv_len;
  int res = lib_sendto(fd, &upkt, send_len, 0, &to_addr,
		       use_tunnel, &numSentDummy, pkt_dest_scid,
		       serv_my_scid);
  if(res != send_len) {
    if(res < 0) { perror("noop_cl.handleData.lib_sendto"); }
    fprintf(stderr, "res=%d, send_len=%d\n", res, send_len);
  }
} // End noop_client_info::fwdPkt()

void noop_client_info::handleDataAck() {
  fwdPkt(data_ack_sock_fd, DATA_ACK_OFFSET);
} // End noop_client_info::handleDataAck()

void noop_client_info::handleSoftState() {
  fwdPkt(state_sock_fd, STATE_OFFSET);
  fprintf(stderr, "s");
} // End noop_client_info::handleSoftState()
