/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to test heart-beat library */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "HB.h"
#include "../util/Util.h"

void usage(char *progname) {
  fprintf(stderr, "Usage: %s <mySCID> <peerSCID> <remote-host>\n", progname);
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 4) { usage(argv[0]); exit(1); }

  unsigned short mySCID = atoi(argv[1]);
  unsigned short peerSCID = atoi(argv[2]);
  struct in_addr to_addr = nslookup(argv[3]);
  HB_t hb(to_addr, peerSCID);

  bool run_listen_thread = false, run_send_thread = false;
  HB::startListenThread(run_listen_thread);
  HB::start_all_sends(run_send_thread, mySCID);

  struct timeval last_print_time;
  gettimeofday(&last_print_time, NULL);

  for(;;) {
    if(!HB::run_listen_thread) {
      struct timeval timeout = { 0, 50*1000 }; // 100ms
      fd_set rfds; FD_ZERO(&rfds); FD_SET(HB::listen_fd, &rfds);
      select(HB::listen_fd+1, &rfds, NULL, NULL, &timeout);
      if(FD_ISSET(HB::listen_fd, &rfds)) {
	HB::recvHelper();
      }
    } else {
      usleep(50*1000);
    }

    if(!HB::run_send_thread) { HB::all_checkSend(); }

    if(now(last_print_time) >= 2000) {
      fprintf(stderr, "peer_status=%d, my_status_peer_view=%d, lastMeas=%f\n",
	      hb.getPeerStatus(), hb.getMyStatusPeerView(),
	      now(hb.getLastMeas()));
      gettimeofday(&last_print_time, NULL);
    }
  }

} // End main()
