/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to test latency measurement library */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "LatMeas.h"
#include "../util/Util.h"

void usage(char *progname) {
  fprintf(stderr, "Usage: %s <mySCID> <peerSCID> <remote-host>\n", progname);
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 4) { usage(argv[0]); exit(1); }

  unsigned short mySCID = atoi(argv[1]);
  unsigned short peerSCID = atoi(argv[2]);
  struct in_addr to_addr = nslookup(argv[3]);
  LatMeas_t L(to_addr, peerSCID);

  bool run_listen_thread = false, run_send_thread = false;
  LatMeas::startListenThread(run_listen_thread);
  LatMeas::start_all_sends(run_send_thread, mySCID);

  struct timeval last_print_time;
  gettimeofday(&last_print_time, NULL);

  for(;;) {
    if(!LatMeas::run_listen_thread) {
      struct timeval timeout = { 0, 50*1000 }; // 100ms
      fd_set rfds; FD_ZERO(&rfds); FD_SET(LatMeas::listen_fd, &rfds);
      select(LatMeas::listen_fd+1, &rfds, NULL, NULL, &timeout);
      if(FD_ISSET(LatMeas::listen_fd, &rfds)) {
	LatMeas::recvHelper();
      }
    } else {
      usleep(50*1000);
    }

    if(!LatMeas::run_send_thread) { LatMeas::all_checkSend(); }

    if(now(last_print_time) >= 2000) {
      fprintf(stderr, "rtt=%f, srtt=%f, lastMeas=%f\n",
	      L.getRTT(), L.getSRTT(), now(L.getLastMeas()));
      gettimeofday(&last_print_time, NULL);
    }
  }

} // End main()
