/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to parse log-file */

#include <iostream>
#include <fstream>

#include <stdio.h>
#include <math.h>

#include "../util/LogLine.h"
#include "../util/Util.h"
#include "../util/JHashtable.h"

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <merged-log-files>\n", progname);
  fprintf(stderr, "(give \"-\" for using stdin)\n");
} // End usage()

struct path_info {
  int pathID;
  unsigned short path_dest_scid;
  unsigned short path_origin_scid;
  unsigned short version;
  JString services;
  JString physPath;
};
typedef struct path_info path_info_t;

struct edge_path_count {
  unsigned short dest_scid;
  unsigned short origin_scid;
  vector<path_info_t> pInfo;
};
typedef struct edge_path_count edge_path_count_t;

struct e2e_entry {
  unsigned long detect_time;
  unsigned short fail_dest_scid;
  unsigned short fail_origin_scid;
  int pathID;
  unsigned short path_dest_scid;
  unsigned short fail_version;

  bool spurious;

  unsigned long notify_time;
  bool alt_initiated;
  unsigned short alt_version;

  unsigned long alt_done_time;
  unsigned long alt_ack_done_time;

  e2e_entry() {
    detect_time = 0;
    fail_dest_scid = fail_origin_scid = LogLine::INVALID_SCID;
    pathID = -1;
    path_dest_scid = LogLine::INVALID_SCID;
    fail_version = 0;

    spurious = false;

    notify_time = 0;
    alt_initiated = false;
    alt_version = 0;

    alt_done_time = alt_ack_done_time = 0;
  } // End e2e_entry::e2e_entry()

  JString toStr() {
    return "pathID=" + long2str(pathID) + " path_dest_scid="
      + long2str(path_dest_scid) + " version=" + long2str(fail_version)
      + " fail_dest_scid=" + long2str(fail_dest_scid)
      + " fail_origin_scid=" + long2str(fail_origin_scid);
  } // End e2e_entry::toStr()
}; // End struct e2e_entry
typedef struct e2e_entry e2e_entry_t;

struct local_entry {
  unsigned long detect_time;
  int pathID;
  unsigned short path_dest_scid;
  unsigned short fail_version;
  unsigned short fail_dest_scid, fail_origin_scid;
  unsigned long redir_time, ack_time;
  bool e2e_initiated;
  unsigned long fail_ack_time;

  local_entry() {
    detect_time = 0;
    pathID = -1;
    path_dest_scid = LogLine::INVALID_SCID;
    fail_version = 0;
    fail_dest_scid = fail_origin_scid = LogLine::INVALID_SCID;
    redir_time = ack_time = 0;
    fail_ack_time = 0;
    e2e_initiated = false;
  } // End local_entry::local_entry()

  JString toStr() {
    return "pathID=" + long2str(pathID) +
      " path_dest_scid=" + long2str(path_dest_scid) +
      " fail_version=" + long2str(fail_version) +
      " fail_dest_scid=" + long2str(fail_dest_scid) +
      " fail_origin_scid=" + long2str(fail_origin_scid);
  } // End local_entry::toStr()
}; // End struct local_entry
typedef struct local_entry local_entry_t;

vector<edge_path_count_t> ep_count;
vector<path_info_t> paths;
vector<e2e_entry_t> e2e_attempts;
vector<unsigned long> spurious_times;

vector<LogLine_t> lines;
ofstream alt_done_fout, redir_fout;
ofstream paths_fout, pcount_fout;
ofstream alt_norm_fout;

void calc_tot() {
  unsigned long total_sent = 0;
  unsigned long total_recd = 0;
  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];
    if((ll.fn == MAIN_FN) && (ll.desc == total_sent_desc)) {
      total_sent += ll.total_sent;
    }
    if((ll.fn == MAIN_FN) && (ll.desc == total_recd_desc)) {
      total_recd += ll.total_recd;
    }
  }
  cerr << "total_sent=" << total_sent << endl;
  cerr << "total_recd=" << total_recd << endl;
  cerr << endl;
} // End calc_tot()

void calc_e2e() {

  // get vector of e2e attempts
  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];

    if((ll.fn != PATCHEDGEE2EHELPER_FN) || (ll.desc != e2e_patch_desc)) {
      continue;
    }

    if(ll.scid == ll.fail_dest_scid) {
      e2e_entry_t e1;
      e1.pathID = ll.pathID;
      e1.path_dest_scid = ll.path_dest_scid;
      e1.fail_version = ll.fail_version;
      e1.fail_dest_scid = ll.fail_dest_scid;
      e1.fail_origin_scid = ll.fail_origin_scid;
      e1.detect_time = ll.time;

      // see if this was a spurious patch attempt
      // find previous line with num_to_patch
      int prev = -1;
      for(unsigned int j = i-1; j >= 0; j--) {
	LogLine_t &prev_ll = lines[j];
	if((prev_ll.scid == ll.scid) &&
	   (prev_ll.fn == PATCHPATHSFROMPEERRC_FN) &&
	   (prev_ll.desc == detect_time_desc)) {
	  prev = j; break;
	}
      }
      if(prev == -1) {
	cerr << "num_to_patch line not found" << endl;
	cerr << ll.toStr() << endl;
      }
      LogLine_t &prev_ll = lines[prev];
      // see if this time matches with the spurious times
      for(unsigned int j = 0; j < spurious_times.size(); j++) {
	if(prev_ll.time == spurious_times[j]) {
	  e1.spurious = true; break;
	}
	if(spurious_times[j] > prev_ll.time) {
	  break; // assume ascending times
	}
      }

      e2e_attempts.push_back(e1);
    }

    if(ll.scid == ll.path_dest_scid) {
      for(int j = e2e_attempts.size()-1; j >= 0; j--) {
	e2e_entry_t &e1 = e2e_attempts[j];
	if((ll.pathID == e1.pathID) &&
	   (ll.path_dest_scid == e1.path_dest_scid) &&
	   (ll.fail_version == e1.fail_version) &&
	   (ll.fail_dest_scid == e1.fail_dest_scid)) {
	  e1.notify_time = ll.time;
	  break;
	}
      }
    }

  } // End for(getting vector of e2e attempts)

  // mark e2e attempts that started off a create-alt-path
  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];

    if((ll.fn != CREATEPATH_FN) || (ll.desc != cp_req_desc)) {
      continue;
    }

    for(unsigned int j = 0; j < e2e_attempts.size(); j++) {
      e2e_entry_t &e1 = e2e_attempts[j];
      if((ll.pathID == e1.pathID) &&
	 (ll.path_dest_scid == e1.path_dest_scid) &&
	 (ll.version == (e1.fail_version+100)) &&
	 (ll.fail_dest_scid == e1.fail_dest_scid) &&
	 (ll.fail_origin_scid == e1.fail_origin_scid)) {
	e1.alt_initiated = true;
	e1.alt_version = ll.version;
	break;
      }
    }

  } // End marking e2e attempts that started off a create-alt-path

  // find alt-done
  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];

    if((ll.fn != CHECKSERVACKS_FN) || (ll.desc != cp_done_desc)) {
      continue;
    }

    for(unsigned int j = 0; j < e2e_attempts.size(); j++) {
      e2e_entry_t &e1 = e2e_attempts[j];
      if(e1.alt_initiated && (e1.pathID == ll.pathID) &&
	 (e1.path_dest_scid == ll.path_dest_scid) &&
	 (e1.alt_version == ll.version)) {
	e1.alt_done_time = ll.time;
	break;
      }
    }

  } // End finding alt-done at src

  // find alt-ack-done
  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];

    if((ll.fn != CHECKACKS_FN) || (ll.desc != alt_done_desc) ||
       (ll.scid != ll.path_dest_scid)) {
      continue;
    }

    for(unsigned int j = 0; j < e2e_attempts.size(); j++) {
      e2e_entry_t &e1 = e2e_attempts[j];
      if(e1.alt_initiated && (e1.pathID == ll.pathID) &&
	 (e1.path_dest_scid == ll.path_dest_scid) &&
	 (e1.alt_version == ll.version)) {
	e1.alt_ack_done_time = ll.time;
	break;
      }
    }
  } // End finding alt-done at dest

  // do computations
  int num_notified = 0;
  int num_alt_initiated = 0;
  int num_spurious_ini = 0;
  int num_alt_done = 0;
  int num_alt_ack_done = 0;
  int num_below_1500 = 0;

  float sum_notify = 0, sum_sq_notify = 0;
  float sum_alt_done = 0, sum_sq_alt_done = 0;
  float sum_alt_ack_done = 0, sum_sq_alt_ack_done = 0;
  float sum_below_1500 = 0, sum_sq_below_1500 = 0;

  for(unsigned int i = 0; i < e2e_attempts.size(); i++) {
    e2e_entry_t &e1 = e2e_attempts[i];
    float delay;

    if(e1.notify_time == 0) {
      //cerr << "Notify fail for " << e1.toStr() << endl;
      continue;
    }
    num_notified++;
    delay = (e1.notify_time - e1.detect_time);
    sum_notify += delay; sum_sq_notify += delay*delay;

    if(!e1.alt_initiated) {
      //cerr << "Alt not initiated for " << e1.toStr() << endl;
      continue;
    }
    num_alt_initiated++;
    if(e1.spurious) { num_spurious_ini++; }
    if(e1.alt_done_time == 0) {
      //cerr << "Alt not done for " << e1.toStr() << endl;
      continue;
    }
    num_alt_done++;
    delay = (e1.alt_done_time - e1.detect_time);
    sum_alt_done += delay; sum_sq_alt_done += delay*delay;
    alt_done_fout << delay << endl;

    if(delay <= 1500) {
      num_below_1500++;
      sum_below_1500 += delay;
      sum_sq_below_1500 += delay*delay;
    }

    if(delay > 3000) {
      //cerr << "Above 3000 for " << e1.toStr() << endl;
    }

    if(e1.alt_ack_done_time == 0) {
      //cerr << "Alt ack not done for " << e1.toStr() << endl;
      continue;
    }
    num_alt_ack_done++;
    delay = (e1.alt_ack_done_time - e1.detect_time);
    sum_alt_ack_done += delay; sum_sq_alt_ack_done += delay*delay;
  }

  float avg_notify = 0, se_notify = 0;
  float avg_alt_done = 0, se_alt_done = 0;
  float avg_alt_ack_done = 0, se_alt_ack_done = 0;
  float avg_below_1500 = 0, se_below_1500 = 0;

  if(num_notified > 0) {
    avg_notify = sum_notify/num_notified;
    se_notify = sqrt(sum_sq_notify/num_notified - avg_notify*avg_notify);
  }

  if(num_alt_done > 0) {
    avg_alt_done = sum_alt_done/num_alt_done;
    se_alt_done = sqrt
      (sum_sq_alt_done/num_alt_done - avg_alt_done*avg_alt_done);
  }

  if(num_below_1500 > 0) {
    avg_below_1500 = sum_below_1500/num_below_1500;
    se_below_1500 = sqrt
      (sum_sq_below_1500/num_below_1500 - avg_below_1500*avg_below_1500);
  }

  if(num_alt_ack_done > 0) {
    avg_alt_ack_done = sum_alt_ack_done/num_alt_ack_done;
    se_alt_ack_done = sqrt(sum_sq_alt_ack_done/num_alt_ack_done -
			   avg_alt_ack_done*avg_alt_ack_done);
  }

  float perc_notified = 100, perc_initiated = 100;
  float perc_spurious_ini = 0;
  float perc_alt_done = 100, perc_alt_ack_done = 100;
  float perc_below_1500 = 100, perc_above_1500 = 0;
  int num_above_1500 = num_alt_initiated;
  if(e2e_attempts.size() > 0) {
    perc_notified = 100.0*num_notified/e2e_attempts.size();
    perc_initiated = 100.0*num_alt_initiated/e2e_attempts.size();
  }
  if(num_alt_initiated > 0) {
    perc_spurious_ini = 100.0*num_spurious_ini/num_alt_initiated;
    perc_alt_done = 100.0*num_alt_done/num_alt_initiated;
    perc_alt_ack_done = 100.0*num_alt_ack_done/num_alt_initiated;
    perc_below_1500 = 100.0*num_below_1500/num_alt_initiated;
    perc_above_1500 = 100.0-perc_below_1500;
    num_above_1500 = num_alt_initiated-num_below_1500;
  }

  cerr << "Total e2e attempts = " << e2e_attempts.size() << endl;
  cerr << "Num notified       = " << num_notified << " ("
       << perc_notified << "%)" << endl;
  cerr << "Num unique e2e     = " << num_alt_initiated << " ("
       << perc_initiated << "%)" << endl;
  cerr << "Num spurious ini   = " << num_spurious_ini << " ("
       << perc_spurious_ini << "%)" << endl;
  cerr << "Num alt done       = " << num_alt_done << " ("
       << perc_alt_done << "%)" << endl;
  cerr << "Num alt ack done   = " << num_alt_ack_done << " ("
       << perc_alt_ack_done << "%)" << endl;
  cerr << "Avg time to notify = " << avg_notify << " ("
       << se_notify << ")" << endl;
  cerr << "Avg alt done       = " << avg_alt_done << " ("
       << se_alt_done << ")" << endl;
  cerr << "Avg alt ack done   = " << avg_alt_ack_done << " ("
       << se_alt_ack_done << ")" << endl;
  cerr << endl;

  cerr << "Num below 1500     = " << num_below_1500 << " ("
       << perc_below_1500 << "%)" << endl;
  cerr << "Num above 1500     = " << num_above_1500 << " ("
       << perc_above_1500 << "%)" << endl;
  cerr << "Avg below 1500     = " << avg_below_1500 << " ("
       << se_below_1500 << ")" << endl;
  cerr << endl;

} // End calc_e2e()

void calc_local() {

  vector<local_entry_t> local_attempts;

  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];

    if((ll.fn == PATCHEDGELOCAL_FN) && (ll.desc == replacement_found_desc)) {
      // look back to get the path details
      int ind = -1;
      for(int j = i-1; j >= 0; j--) {
	LogLine_t &prev_ll = lines[j];
	if((prev_ll.scid == ll.scid) &&
	   (prev_ll.fn == PATCHEDGELOCAL_FN) &&
	   (prev_ll.desc == try_local_desc)) {
	  ind = j; break;
	}
      }

      if(ind < 0) {
	cerr << "try-local not found" << endl;
	cerr << ll.toStr() << endl;
	exit(9);
      }
      LogLine_t &prev_ll = lines[ind];

      // add local attempt entry into list
      local_entry_t local;
      local.detect_time = prev_ll.time;
      local.pathID = prev_ll.pathID;
      local.path_dest_scid = prev_ll.path_dest_scid;
      local.fail_version = prev_ll.version;
      local.fail_dest_scid = prev_ll.scid;
      local.fail_origin_scid = prev_ll.fail_origin_scid;
      local_attempts.push_back(local);
    } // End handling patchEdgeLocal

    if((ll.fn == CHANGESERVINST_FN) && (ll.desc == csi_req_desc)) {
      // find the previous line in the merged log file
      int ind = -1;
      for(int j = i-1; j >= 0; j--) {
	LogLine_t &prev_ll = lines[j];
	if((prev_ll.scid == ll.scid) &&
	   (prev_ll.fn == PEERPATCHPATHLOCAL_FN) &&
	   (prev_ll.desc == req_desc)) {
	  ind = j; break;
	}
      }

      if(ind < 0) {
	cerr << "peerPatchPathLocal not found" << endl;
	cerr << ll.toStr() << endl;
	exit(5);
      }
      LogLine_t &prev_ll = lines[ind];

      if(prev_ll.scid != prev_ll.fail_origin_scid) {
	cerr << "csi-req at non-end-point?" << endl;
	cerr << prev_ll.toStr() << endl;
      }

      // find the matching local recovery attempt
      int att_num = -1;
      for(int j = local_attempts.size()-1; j >= 0; j--) {
	local_entry_t &local = local_attempts[j];
	if((local.fail_origin_scid == prev_ll.scid) &&
	   (local.fail_dest_scid == prev_ll.fail_dest_scid) &&
	   (local.pathID == prev_ll.pathID) &&
	   (local.path_dest_scid == prev_ll.path_dest_scid) &&
	   (local.fail_version == prev_ll.fail_version)) {
	  att_num = j; break;
	}
      }

      if(att_num < 0) {
	cerr << "no matching local attempt for peerPatchPathLocal" << endl;
	cerr << ll.toStr() << endl;
	exit(6);
      }
      local_attempts[att_num].redir_time = ll.time;
    } // End handling changeServInst, peerPatchPathLocal

    if((ll.fn == CHECKACKS_FN) && (ll.desc == repair_done_desc) &&
       (ll.scid == ll.fail_dest_scid)) {
      // find the matching local recovery attempt
      int att_num = -1;
      for(int j = local_attempts.size()-1; j >= 0; j--) {
	local_entry_t &local = local_attempts[j];
	if((local.fail_dest_scid == ll.scid) &&
	   (local.fail_origin_scid == ll.fail_origin_scid) &&
	   (local.pathID == ll.pathID) &&
	   (local.path_dest_scid == ll.path_dest_scid) &&
	   (local.fail_version == ll.version)) {
	  att_num = j; break;
	}
      }

      if(att_num < 0) {
	cerr << "no matching local attempt for checkAcks" << endl;
	cerr << ll.toStr() << endl;
	exit(6);
      }
      local_attempts[att_num].ack_time = ll.time;
    } // end handling checkAcks, repair-done

    if((ll.fn == CHECKACKS_FN) && (ll.desc == rcv_fail_desc)) {
      // find the next entry in the log-file with the same scid
      unsigned int ind = lines.size();
      for(unsigned int j = i+1; j < lines.size(); j++) {
	LogLine_t &next_ll = lines[j];
	if(next_ll.scid == ll.scid) { ind = j; break; }
      }
      if(ind == lines.size()) {
	cerr << "nothing after rcv-fail" << endl;
	cerr << ll.toStr() << endl;
	exit(10);
      }

      LogLine_t &next_ll = lines[ind];
      if((next_ll.fn == CHECKACKS_FN) &&
	 (next_ll.desc == local_fail_try_e2e_desc)) {
	// find the next e2e line
	unsigned int ind2 = lines.size();
	for(unsigned int j = ind+1; j < lines.size(); j++) {
	  LogLine_t &next_ll2 = lines[j];
	  if(next_ll2.scid == ll.scid) { ind2 = j; break; }
	}

	// sanity checks
	if(ind2 == lines.size()) {
	  cerr << "can't find e2e after local-fail-try-e2e" << endl;
	  cerr << next_ll.toStr() << endl;
	  exit(11);
	}
	LogLine_t &next_ll2 = lines[ind2];
	if((next_ll2.fn != PATCHEDGEE2EHELPER_FN) ||
	   (next_ll2.desc != e2e_patch_desc) ||
	   (next_ll2.scid != next_ll2.fail_dest_scid)) {
	  cerr << "line is not e2e-patch after local-fail-try-e2e" << endl;
	  cerr << next_ll2.toStr() << endl;
	  exit(12);
	}

	// find the matching local attempt entry and mark it
	int l_ind = -1;
	for(int j = local_attempts.size()-1; j >= 0; j--) {
	  local_entry_t &local = local_attempts[j];
	  if((local.pathID == next_ll2.pathID) &&
	     (local.path_dest_scid == next_ll2.path_dest_scid) &&
	     (local.fail_version == next_ll2.fail_version) &&
	     (local.fail_dest_scid == next_ll2.fail_dest_scid) &&
	     (local.fail_origin_scid == next_ll2.fail_origin_scid)) {
	    l_ind = j; break;
	  }
	}
	if(l_ind < 0) {
	  cerr << "no matching local entry for e2e" << endl;
	  cerr << next_ll2.toStr() << endl;
	  exit(13);
	}

	local_attempts[l_ind].e2e_initiated = true;
	local_attempts[l_ind].fail_ack_time = ll.time;
      }

    } // end handling checkAcks, rcv-fail

  } // end going through all the lines

  // calculate and print statistics
  int num_redir = 0, num_e2e_ini = 0, num_ack = 0;
  float sum_redir = 0, sum_ack = 0;
  float sum_sq_redir = 0, sum_sq_ack = 0;

  for(unsigned int i = 0; i < local_attempts.size(); i++) {
    local_entry_t &local = local_attempts[i];
    float delay;

    if(local.e2e_initiated) {
      num_e2e_ini++;
      continue;
    }

    if(local.redir_time == 0) {
      //cerr << "Local redir failed for " << local.toStr() << endl;
      continue;
    }
    num_redir++;

    delay = local.redir_time - local.detect_time;
    sum_redir += delay; sum_sq_redir += delay*delay;
    redir_fout << delay << endl;

    if(local.ack_time == 0) {
      //cerr << "Local ack failed for " << local.toStr() << endl;
      continue;
    }
    num_ack++;

    delay = local.ack_time - local.detect_time;
    sum_ack += delay; sum_sq_ack += delay*delay;
  }

  float avg_redir = 0, avg_ack = 0;
  float se_redir = 0, se_ack = 0;
  if(num_redir > 0) {
    avg_redir = sum_redir/num_redir;
    se_redir = sqrt(sum_sq_redir/num_redir - avg_redir*avg_redir);
  }
  if(num_ack > 0) {
    avg_ack = sum_ack/num_ack;
    se_ack = sqrt(sum_sq_ack/num_ack - avg_ack*avg_ack);
  }

  float perc_redir = 100, perc_ack = 100, perc_e2e_ini = 0;
  if(local_attempts.size() > 0) {
    perc_redir = 100*num_redir/local_attempts.size();
    perc_ack = 100*num_ack/local_attempts.size();
    perc_e2e_ini = 100*num_e2e_ini/local_attempts.size();
  }

  cerr << "Total local recovery attempts  = " << local_attempts.size()
       << endl;
  cerr << "Number local fail, e2e ini     = " << num_e2e_ini << " ("
       << perc_e2e_ini << "%)" << endl;
  cerr << "Number redirected successfully = " << num_redir << " ("
       << perc_redir << "%)" << endl;
  cerr << "Number acked on redirection    = " << num_ack << " ("
       << perc_ack << "%)" << endl;
  cerr << "Avg time to redirect           = " << avg_redir << " ("
       << se_redir << ")" << endl;
  cerr << "Avg time to get ack            = " << avg_ack << " ("
       << se_ack << ")" << endl;
  cerr << endl;

} // End calc_local()

void calc_plen() {
  float sum_len = 0, sum_sq_len = 0;
  int numPaths = 0;
  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];
    if((ll.fn != CREATEPATH_FN) || (ll.desc != phys_path_found_desc)) {
      continue;
    }
    const JString &physPathStr = ll.strip_getValue("physPath");
    vector<JString> pVec;
    int pLen = split(physPathStr, pVec);
    sum_len += pLen;
    sum_sq_len += pLen*pLen;
    numPaths++;
  }

  float avg_len = 0, se_len = 0;
  if(numPaths > 0) {
    avg_len = sum_len/numPaths;
    se_len = sqrt(sum_sq_len/numPaths - avg_len*avg_len);
  }

  cerr << "Total number of paths = " << numPaths << endl;
  cerr << "Average path length   = " << avg_len << " ("
       << se_len << ")" << endl;
  cerr << endl;
} // End calc_plen()

void add_path_to_edge(const path_info_t &path,
		      unsigned short e_dest, unsigned short e_origin) {

  int ind = -1;
  for(unsigned int i = 0; i < ep_count.size(); i++) {
    if((e_dest == ep_count[i].dest_scid) &&
       (e_origin == ep_count[i].origin_scid)) {
      ind = i; break;
    }
  }
  if(ind == -1) {
    ind = ep_count.size();
    edge_path_count_t ep;
    ep.dest_scid = e_dest;
    ep.origin_scid = e_origin;
    ep_count.push_back(ep);
  }

  edge_path_count_t &ep = ep_count[ind];

  // check if path has already been added here
  for(unsigned int j = 0; j < ep.pInfo.size(); j++) {
    path_info_t &p = ep.pInfo[j];
    if((p.pathID == path.pathID) && (p.path_dest_scid == path.path_dest_scid) &&
       (p.path_origin_scid == path.path_origin_scid)) {
      return;
    }
  }
  ep.pInfo.push_back(path);

} // End add_path_to_edge()

char alt_done_file[256], redir_file[256], paths_file[256], pcount_file[256];
char alt_norm_file[256];

void calc_path_per_edge() {

  for(unsigned int i = 0; i < lines.size(); i++) {
    LogLine_t &ll = lines[i];
    if((ll.fn != CREATEPATH_FN) || (ll.desc != cp_req_desc)) {
      continue;
    }

    // find next entry of same scid
    unsigned int ind = lines.size();
    for(unsigned int j = i+1; j < lines.size(); j++) {
      LogLine_t &next_ll = lines[j];
      if(next_ll.scid == ll.scid) {
	if((next_ll.fn == CREATEPATH_FN) &&
	   (next_ll.desc == graph_state_stale_desc)) {
	  continue;
	}
	if((next_ll.fn == LISTENTHREAD_FN) ||
	   (next_ll.fn == SENDTHREAD_FN)) {
	  continue;
	}
	ind = j; break;
      }
    }
    if(ind == lines.size()) { continue; }

    // sanity check
    LogLine_t &next_ll = lines[ind];
    if((next_ll.fn != CREATEPATH_FN) || (next_ll.desc != phys_path_found_desc)) {
      continue;
    }

    const JString &pStr = next_ll.strip_getValue("physPath");

    vector<JString> pVec;
    int pLen = split(pStr, pVec);
    vector<unsigned short> scidVec;

    for(int j = 0; j < pLen; j++) {
      char buf[128];
      JString_cp(buf, pVec[j]);
      char *col_pos = index(buf, ':');
      if(col_pos == NULL) {
	cerr << "colon not found in physPath buf=" << buf << endl;
	cerr << next_ll.toStr() << endl;
	exit(35);
      }
      *col_pos = 0;
      unsigned short scid = atoi(buf);
      scidVec.push_back(scid);
    }

    // add path info to list
    path_info_t p;
    p.pathID = ll.pathID; p.path_dest_scid = ll.path_dest_scid;
    p.path_origin_scid = ll.path_origin_scid;
    p.services = ll.strip_getValue("services");
    p.version = ll.version;
    p.physPath = pStr;
    paths.push_back(p);

    if(ll.version != 100) { continue; }

    for(unsigned int j = 0; j < scidVec.size()-1; j++) {
      if(scidVec[j] != scidVec[j+1]) {
	add_path_to_edge(p, scidVec[j], scidVec[j+1]);
	add_path_to_edge(p, scidVec[j+1], scidVec[j]);
      }
    }

    // print to paths_fout file
    paths_fout << ll.toStr() << " physPath=[" << pStr << ']' << endl;

  } // End for(reading all lines in log)

  // print to pcount file, after sorting according to path count
  for(unsigned int i = 0; i < ep_count.size(); i++) {
    for(unsigned int j = i+1; j < ep_count.size(); j++) {
      if(ep_count[j].pInfo.size() < ep_count[i].pInfo.size()) {
	edge_path_count_t temp = ep_count[i];
	ep_count[i] = ep_count[j];
	ep_count[j] = temp;
      }
    }
  }
  for(unsigned int i = 0; i < ep_count.size(); i++) {
    pcount_fout << ep_count[i].dest_scid << '\t' << ep_count[i].origin_scid
		<< '\t' << ep_count[i].pInfo.size() << endl;
  }

} // End calc_path_per_edge()

void calc_norm_e2e() {
  int norm_count = 0;
  float sum_norm = 0, sum_sq_norm = 0;

  for(unsigned int i = 0; i < e2e_attempts.size(); i++) {
    e2e_entry_t &e1 = e2e_attempts[i];

    if((e1.notify_time == 0) || (!e1.alt_initiated) ||
       (e1.alt_done_time == 0)) {
      continue;
    }

    // find corr path in vector of paths
    int ind = -1;
    for(unsigned int j = 0; j < paths.size(); j++) {
      path_info_t &p = paths[j];
      if((e1.pathID == p.pathID) && (e1.path_dest_scid == p.path_dest_scid) &&
	 (e1.fail_version == p.version)) {
	ind = j; break;
      }
    }

    if(ind == -1) {
      cerr << "calc_norm_e2e: recovered path not found p_id="
	   << e1.pathID << ':' << e1.path_dest_scid 
	   << " path_ver=" << e1.fail_version << endl;
      exit(4);
    }

    path_info_t &p = paths[ind];
    vector<JString> pVec;
    int physLen = split(p.physPath, pVec);

    if(physLen <= 0) {
      cerr << "weird: physLen=" << physLen << endl;
      exit(5);
    }

    int notify_len = 0;
    for(unsigned int j = 0; j < pVec.size(); j++) {
      char buf[128];
      JString_cp(buf, pVec[j]);
      char *col_pos = index(buf, ':');
      if(col_pos == NULL) {
	cerr << "colon not found in physPath buf=" << buf << endl;
	exit(35);
      }
      *col_pos = 0;
      unsigned short scid = atoi(buf);
      if(scid == e1.fail_dest_scid) { notify_len = j; break; }
    }

    float delay = (e1.alt_done_time - e1.detect_time);
    float norm_delay = delay/(physLen+notify_len);
    sum_norm += norm_delay;
    sum_sq_norm += norm_delay*norm_delay;
    norm_count++;

    alt_norm_fout << norm_delay << endl;

  }

  float avg_norm = 0, se_norm = 0;
  if(norm_count > 0) {
    avg_norm = sum_norm/norm_count;
    se_norm = sqrt(sum_sq_norm/norm_count - avg_norm*avg_norm);
  }

  cerr << "Norm avg alt done  = " << avg_norm << " ("
       << se_norm << ")" << endl;
  cerr << endl;

} // End calc_norm_e2e()

void get_spurious() {
  ifstream fin("spurious-times.out");
  if(fin.fail()) {
    cerr << "Warning: Can't open " << "spurious-times.out" << endl;
    return;
  }
  for(;;) {
    // read each line
    if(fin.eof()) break;
    char line[2048];
    fin.getline(line, sizeof(line)); // read a line
    if(fin.eof()) break;
    if(fin.fail()) {
      cerr << "Error reading" << endl; exit(99);
    }
    unsigned long time = atoi(line);
    spurious_times.push_back(time);
  }
  fin.close();
} // End get_spurious()

void main_helper(istream &fin) {

  // output file of alt-done measurements
  alt_done_fout.open(alt_done_file);
  if(alt_done_fout.fail()) {
    fprintf(stderr, "Can't write to file %s\n", alt_done_file);
    exit(45);
  }

  // output file of local redirection times
  redir_fout.open(redir_file);
  if(redir_fout.fail()) {
    fprintf(stderr, "Can't write to file %s\n", redir_file);
    exit(46);
  }

  // output file of paths
  paths_fout.open(paths_file);
  if(paths_fout.fail()) {
    fprintf(stderr, "Can't write to file %s\n", paths_file);
    exit(45);
  }

  // output file of path counts
  pcount_fout.open(pcount_file);
  if(pcount_fout.fail()) {
    fprintf(stderr, "Can't write to file %s\n", pcount_file);
    exit(45);
  }

  // output file of normalized e2e recovery times
  alt_norm_fout.open(alt_norm_file);
  if(alt_norm_fout.fail()) {
    fprintf(stderr, "Can't write to file %s\n", alt_norm_file);
    exit(45);
  }

  // read the stuff into memory
  for(;;) {
    // read each line
    if(fin.eof()) break;
    char line[2048];
    fin.getline(line, sizeof(line)); // read a line
    if(fin.eof()) break;
    if(fin.fail()) {
      cerr << "Error reading" << endl; exit(99);
    }
    LogLine_t ll(line);
    if(!ll.succ) {
      cerr << "Format wrong while reading line: " << lines.size() << endl;
      cerr << line << endl;
      exit(9);
    }
    if((lines.size() % 10000) == 0) {
      cerr << "(l=" << lines.size() << ')';
    }
    lines.push_back(ll);
  }
  cerr << endl;

  get_spurious();
  calc_tot();
  calc_e2e();
  calc_local();
  calc_plen();
  if(!cfg_opt_logline) {
    calc_path_per_edge();
    calc_norm_e2e();
  }

} // End main_helper

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 2) { usage(argv[0]); exit(1); }
  const char *log_file = argv[1];

  if(strcmp(log_file, "-") == 0) {
    sprintf(alt_done_file, "alt-done.out");
    sprintf(redir_file, "redir.out");
    sprintf(paths_file, "paths.out");
    sprintf(pcount_file, "pcount.out");
    sprintf(alt_norm_file, "alt-norm.out");
    main_helper(cin);
  } else {
    sprintf(alt_done_file, "alt-done-%s", log_file);
    sprintf(redir_file, "redir-%s", log_file);
    sprintf(paths_file, "paths-%s", log_file);
    sprintf(pcount_file, "pcount-%s", log_file);
    sprintf(alt_norm_file, "alt-norm-%s", log_file);
    ifstream fin(log_file);
    if(fin.fail()) {
      cerr << "Can't open " << log_file << endl;
      exit(2);
    }
    main_helper(fin);
  }

} // End main()
