/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef SERV_COMP_IF_PKT_H
#define SERV_COMP_IF_PKT_H

#include <vector>

#include "../util/LogLine.h"

enum ServCompPktType {
  INVALID_TYPE,
  APP_CREATE_PATH, APP_KILL_PATH, APP_FWD_DATA,

  PEER_CREATE_SERV_INST, PEER_KILL_SERV_INST,
  PEER_FWD_DATA, PEER_PATCH_PATH, PEER_ALT_PATH,
  PEER_PATH_HB, PEER_PATCH_BUNCH,

  APP_CREATE_SERV_INST, APP_KILL_SERV_INST,
  APP_CHANGE_SERV_INST,

  APP_UPSTREAM_INFO, // give info about appln upstream
  APP_PATH_STATE, // to pass around appln soft state
  APP_HB, // sent from one app to upstream app
  APP_KILL, // sent to kill upstream path

  TTS_REQ, // request for text
  TTS_RESP, // response consisting of text

};
typedef enum ServCompPktType ServCompPktType_t;

typedef struct pkt_type_name {
  ServCompPktType_t type;
  char name[64];
} pkt_type_name_t;

enum StrKey {
  INVALID_F,
  // for path creation
  PREV_DEST_SPEC_F, SERVICES_F, CREATE_SERV_F, DOWNSTREAM_PATH_F,
  // for path repair
  OLD_PATH_F, NEW_SUB_PATH_F,
  // for apps
  UPSTREAM_SPEC_F, PATH_STATE_F,
  // for tts app
  REQ_ARGS_F, RESP_TEXT_F,
};
typedef enum StrKey StrKey_t;

struct StrField {
  int16_t key;
  u_int16_t len;
  char value[1024];
  StrField();
  StrField(const StrField &sf);
};
typedef struct StrField StrField_t;

struct FailedPath {
  int32_t pathID;
  u_int16_t path_dest_scid;
  u_int16_t version;
};
typedef struct FailedPath FailedPath_t;

enum RepairType { LOCAL_R, E2E_R };
typedef enum RepairType RepairType_t;

struct FixedPkt {
  /** Packet type */
  int16_t type;

  /** Path info */
  int32_t pathID;
  int32_t app_seq, port;
  u_int16_t path_dest_scid, path_origin_scid;
  u_int16_t version;

  /** Path traversal info */
  int16_t posn, subPosn;

  /** Repair info */
  int16_t repType;
  u_int16_t fail_dest_scid, fail_origin_scid;
  u_int16_t fail_version;
  int16_t failPosn;

  /** Failure notification */
  u_int16_t failure_scid;
  int16_t succSoFar;
  int16_t numDestroyed;

  /** Communication with services */
  u_int16_t serv_dest_scid, serv_my_scid, serv_upstream_scid;

  /** Success indication */
  int16_t success;

  /** The number of bunched failure info */
  u_int16_t num_in_bunch;
  /** The number of string fields */
  u_int16_t numStr;

  /** The constructor */
  FixedPkt();
};
typedef struct FixedPkt FixedPkt_t;

/** Class to handle the interface packets that the serice-composition
    layer sends/receives */
class ServCompIFPkt {
protected:
  /** For printing purposes */
  static pkt_type_name_t pkt_type_name_map[];
public:
  /** The fixed part of the packet */
  FixedPkt_t fixed;

  /** For bunching */
  vector<FailedPath_t> failed;

  /** Key-value pairs consisting of StrKey_t keys */
  vector<StrField_t> strFields;

  /** Parse the given packet and put data into 'this' object.  The
      length of the data is also given.  Returns 1 on success, 0 on
      failure */
  int parsePkt(char *data, int len);
  /** Print the packet into the given buffer, and return the length of
      the buffer.  The maximum size of the 'data' is also given */
  int toStr(char *data, int maxLen);
  /** Return the length as would be returned by toStr (useful for
      checking size of data) */
  int totLen();

  /** Add a key-value pair to the pkt */
  void addPair(const StrKey_t key, const char *value);

  /** Get the value corresponding to the given key */
  const char *getValue(const StrKey_t key);

  /** For printing purposes */
  static const char *type2str(ServCompPktType_t type);

}; // End class ServCompIFPkt
typedef class ServCompIFPkt ServCompIFPkt_t;

#endif
