/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements TreeInfo.h */

#include "TreeInfo.h"
#include "../util/Util.h"

JHashtable<JString, TreeInfo_t*> TreeInfo::tHash;
float TreeInfo::TREE_CACHE_TIMEOUT = 100*1000; // in ms
bool TreeInfo::debug = false;

void TreeInfo::allocSeenTreeNodes() {
  seenTreeNodes =
    new PriorityQueue<unsigned short>[services.size()+1](numTreeNodes);
  if(seenTreeNodes == NULL) {
    cerr << "Cannot allocate for seenTreeNodes" << endl;
    exit(92);
  }
} // End TreeInfo::allocSeenTreeNodes()

void TreeInfo::deallocSeenTreeNodes() {
  delete[] seenTreeNodes;
} // End TreeInfo::deallocSeenTreeNodes()

void TreeInfo::buildReverseIndex() {
  for(int i = 0; i < numTreeNodes; i++) {
    scidHash.put(nVec[i].scid, i);
  }
} // End TreeInfo::buildReverseIndex()

int TreeInfo::getIndexOfSCID(unsigned short scid) {
  if(scidHash.containsKey(scid)) {
    return scidHash.get(scid);
  } else {
    return -1;
  }
} // End TreeInfo::getIndexOfSCID()

TreeInfo_t *TreeInfo::getTreeInfo(const vector<JString> &services) {
  // need to set this only once, but no harm...
  tHash.setHashFunction(JStringHash);

  JString joint_str = join(services);
  if(tHash.containsKey(joint_str)) {
    if(TreeInfo::debug) {
      cerr << "found tree for " << joint_str << endl;
    }
    return tHash.get(joint_str);
  } else {
    if(TreeInfo::debug) {
      cerr << "could not find tree for " << joint_str << endl;
    }
    return NULL;
  }
} // End TreeInfo::getTreeInfo()

TreeInfo_t *TreeInfo::makeNewTree(const vector<JString> &services,
				  int numTreeNodes) {
  // need to set this only once, but no harm...
  tHash.setHashFunction(JStringHash);

  JString joint_str = join(services);
  if(tHash.containsKey(joint_str)) {
    TreeInfo_t *t = tHash.get(joint_str);
    if(t != NULL) {
      if(TreeInfo::debug) {
	cerr << "destroying tree for " << joint_str << endl;
      }
      delete t; // delete the old tree
    }
  }
  // allocate for new tree 
  TreeInfo_t *newTree = new TreeInfo(services, numTreeNodes);
  if(TreeInfo::debug) {
    cerr << "making tree for " << joint_str << endl;
  }
  tHash.put(joint_str, newTree);
  return newTree;
} // End TreeInfo::makeNewTree()
