/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/* Program to evaluate routing in the overlay network */

#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>

extern "C" {
#include <gb_graph.h>
#include <gb_save.h>
}

#include "../util/Util.h"
#include "../util/JHashtable.h"

extern "C" Graph *c_ver_restore_graph(const char *fName);

void usage(char *progname) {
  fprintf(stderr, "Usage: %s <graph-file>\n", progname);
} // End usage()

int main(int argc, char *argv[]) {

  if(argc != 2) {
    usage(argv[0]); exit(1);
  }

  char *graph_file = argv[1];
  Graph *g = c_ver_restore_graph(graph_file);

  if(g == NULL) {
    fprintf(stderr, "Error reading graph; panic_code = %d\n", panic_code);
    exit(2);
  }

  Vertex *v;
  Arc *a;

  JHashtable<JString, int> nodeHash;
  nodeHash.setHashFunction(JStringHash);
  int nodeNum = 1;

  fprintf(stdout, "graph [\n");
  fprintf(stdout, "  directed 0\n");

  for(v = g->vertices; v < g->vertices + g->n; v++) {
    fprintf(stdout, "  node [\n");
    fprintf(stdout, "    label %d\n", nodeNum);
    fprintf(stdout, "    id %d\n", nodeNum);
    fprintf(stdout, "    graphics [\n");
    fprintf(stdout, "      width 2.0\n");
    fprintf(stdout, "      height 2.0\n");
    fprintf(stdout, "      depth 0.0\n");
    fprintf(stdout, "    ]\n");
    fprintf(stdout, "    vgj [\n");
    fprintf(stdout, "      labelPosition \"Below\"\n");
    fprintf(stdout, "      shape \"Rectangle\"\n");
    fprintf(stdout, "    ]\n");
    fprintf(stdout, "  ]\n");
    nodeHash.put(v->name, nodeNum);
    nodeNum++;
  }

  for(v = g->vertices; v < g->vertices + g->n; v++) {
    long v1 = ((long)(v) - (long)(g->vertices))/sizeof(*v);
    int n1 = nodeHash.get(v->name);
    for(a = v->arcs; a; a = a->next) {
      long v2 = ((long)(a->tip) - (long)(g->vertices))/sizeof(*v);
      int n2 = nodeHash.get(a->tip->name);
      if(v1 < v2) {
	// For gml format, the source node should have occurred first
	// in the list
	fprintf(stdout, "  edge [\n");
	fprintf(stdout, "    label \"%d\"\n", a->len);
	fprintf(stdout, "    source %d\n", n1);
	fprintf(stdout, "    target %d\n", n2);
	fprintf(stdout, "  ]\n");
      }
    }
  }

  fprintf(stdout, "]\n");

} // End main()
