/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to convert SGB graph format to linked-list format (as
    expected by NetworkGraph */

#include <stdio.h>

#include "../util/Util.h"
#include "../util/JHashtable.h"

extern "C" {
#include <gb_graph.h>
#include <gb_save.h>
}

extern "C" Graph *c_ver_restore_graph(const char *fName);

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <sgb-graph-file> <ll-output-file>\n",
	  progname);
} // End usage()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 3) { usage(argv[0]); exit(1); }

  const char *sgbFile = argv[1];
  const char *llFile = argv[2];

  // read the input file
  fprintf(stderr, "Reading graph from %s...", sgbFile);
  Graph *g = c_ver_restore_graph(sgbFile);
  if(g == NULL) {
    fprintf(stderr, "Error reading graph; panic_code = %ld\n", panic_code);
    exit(2);
  }
  fprintf(stderr, "done\n");

  // prepare the output file stream
  FILE *fout = fopen(llFile, "w");
  if(fout == NULL) {
    fprintf(stderr, "Can't open %s for writing\n", llFile);
    perror("fopen");
    exit(3);
  }

  Vertex *v;
  Arc *a;

  // get a hashtable of the nodes
  fprintf(stderr, "Building hashtable of nodes...");
  JHashtable<JString, int> nodeHash;
  nodeHash.setHashFunction(JStringHash);
  int nodeNum = 1;
  for(v = g->vertices; v < g->vertices + g->n; v++) {
    nodeHash.put(v->name, nodeNum);
    nodeNum++;
  }
  fprintf(stderr, "done\n");

  // print the graph in linked-list format
  fprintf(stderr, "Printing graph onto %s...", llFile);
  for(v = g->vertices; v < g->vertices + g->n; v++) {
    int n1 = nodeHash.get(v->name);
    fprintf(fout, "%d ", n1);
    for(a = v->arcs; a; a = a->next) {
      int n2 = nodeHash.get(a->tip->name);
      fprintf(fout, "%d %ld ", n2, a->len);
    }
    fprintf(fout, "\n");
  }
  fprintf(stderr, "done\n");

  fclose(fout);

} // End main()
