/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef APP_PKT_H
#define APP_PKT_H

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>

/** Stores local service information */
struct LocalServInfo {
  /** The name of the service */
  char serviceName[16];
  /** The IP address of the service */
  struct in_addr serviceIP;
  /** The port used for communication with the service-info mgt layer */
  unsigned short servicePort;

  /** Convert this into a string for sending over the network (writes
      onto the given data area).  Format is:
      "serviceName:serviceIP:servicePort\n".  This routine returns the
      string length. */
  int toStr(char *dest) {
    sprintf(dest, "%s:%s:%hd\n", serviceName, inet_ntoa(serviceIP),
	    servicePort);
    return strlen(dest);
  } // End LocalServInfo::toStr()

  /** Parse info from a string recd over the network.  Returns 1 on
      successful parsing, 0 on failure.  Format is:
      "<serviceName>:<serviceIP>:<servicePort>" */
  int parsePkt(char *data);
}; // End struct LocalServInfo

typedef struct LocalServInfo LocalServInfo_t;

/** The structure for storing global service information */
struct GlobalServInfo {
  /** The service cluster where the service is present */
  unsigned short scid;
  /** The port used for communication with the service-info mgt layer */
  unsigned short servicePort;
  /** The IP address of the service */
  struct in_addr serviceIP;
}; // End struct GlobalServInfo

typedef struct GlobalServInfo GlobalServInfo_t;

/** The class that defines the interface packets for application <->
    service-info exchange */
class AppPkt {
public:
  static const char REGISTER;
  static const char DEREGISTER;

  /** One of REGISTER, DEREGISTER */
  char type;
  /** The service information */
  LocalServInfo_t lServInfo;

  /** Parse a packet received from an application.  Return 1 on
      success, 0 on failure.  Format of 'data' is:
      register packet:
      "register:<serviceName>:<serviceIP>:<servicePort>\n"
      deregister packet:
      "deregister:<serviceName>:<serviceIP>:<servicePort>\n"
      <serviceIP> has to be an IP address, not a host name */
  int parsePkt(char *data, int len);
}; // End class AppPkt

typedef class AppPkt appPkt_t;

#endif
