/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements tts_client_info.h */

#include <iostream>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>

#include "tts_client_info.h"
#include "../util/Util.h"
#include "../udp-lib/UDPLib.h"
#include "../serv-comp/ServCompIFPkt.h"

thr_fn_t tts_client_info::thr_fn;
long tts_client_info::SOFT_STATE_PERIOD = 1000; // 1sec

void tts_client_info::init_fields() {
  generic_client_info::init_fields();
  currSentenceSend = currSentenceProc = currSentenceReq = 0;
  allSentencesDone = false;
  startSentence = 0; startBytePosn = 0;
  first_send = true;
  seqno = 0; timestamp = 0;
  fmt_int = RTP_PT_PCMU; send_size = 160; send_interval = 20;
  while(aTTS_vec.size() > 0) { aTTS_vec.pop_back(); }
} // End tts_client_info::init_fields()

void tts_client_info::setThreadFn(thr_fn_t fn) {
  thr_fn = fn;
} // End tts_client_info::setThreadFn()

void tts_client_info::handleData() {
  // recv the pkt
  UDPLibPkt_t uPkt; ServCompIFPkt_t sPkt;
  int recv_len; struct sockaddr_in from_addr;
  unsigned long rcv_count = 0; bool pkt_ready = true;
  int res = UDPLib::blockingRecv(data_sock_fd, &uPkt, &recv_len,
				 &from_addr, NULL, &rcv_count, pkt_ready);
  if(res != 1) { cerr << "res = " << res << endl; return; }
  // parse it
  int parseRes = sPkt.parsePkt(uPkt.data, recv_len);
  if(parseRes != 1) { cerr << "parseRes = " << parseRes << endl; return; }

  JString err_str;
  if(!recordArgs(&sPkt, err_str)) {
    cerr << "record args error" << endl;
    dbg->print() << name << ' ' << err_str << " p_id=" << pathID << ':'
		 << path_dest_scid << " path_ver=" << version << " type="
		 << ServCompIFPkt::type2str((ServCompPktType_t)sPkt.fixed.type)
		 << endl;
    dbg->close();
    exit(85);
  }
  cerr << "startSentence=" << startSentence << " startBytePosn="
       << startBytePosn << endl;

  // send response
  ServCompIFPkt_t resp_sPkt;
  resp_sPkt.fixed.success = 1;
  int send_len = resp_sPkt.toStr(uPkt.data, sizeof(uPkt.data));
  int send_fd = data_ack_sock_fd; // can use same sock to send pkts too
  // should send pkt via the delaying tunnel
  bool use_tunnel = ((serv_dest_scid != serv_my_scid) &&
		     (serv_dest_scid != LogLine::INVALID_SCID));
  unsigned long count = 0;
  res = UDPLib::sendResp(send_fd, &uPkt, send_len,
			 from_addr.sin_addr, destPort+DATA_ACK_OFFSET,
			 use_tunnel, &count);
  if(res != 1) {
    cerr << "cl.handleData: res = " << res << endl; return;
  }

  currSentenceSend = currSentenceProc = currSentenceReq = startSentence;
  while(aTTS_vec.size() < currSentenceReq) {
    asyncTTS_t temp_aTTS; aTTS_vec.push_back(temp_aTTS);
  }
  sendDataReq(); // send request to 'email' service upstream
} // End tts_client_info::handleData()

bool tts_client_info::recordArgs(ServCompIFPkt_t *sPkt, JString &err_str) {
  // sanity checks
  if(sPkt->fixed.type != TTS_REQ) {
    err_str = "unknown-pkt-type"; return false;
  }
  const char *state_str = sPkt->getValue(REQ_ARGS_F);
  if(state_str == NULL) { err_str = "no-state-str"; return false; }

  // get state info
  char *col_pos = index(state_str, ':');
  if(col_pos == NULL) { err_str = "no-col"; return false; }
  *col_pos = 0;
  startSentence = atoi(state_str);
  startBytePosn = atoi(col_pos+1);
  err_str = "ok"; return true;
} // End tts_sink_cl::recordArgs()

void tts_client_info::handleDataAck() {
  int reqNo_recd = rrBlock->check_one(data_ack_sock_fd);
  if(reqNo_recd < 0) { cerr << "[t:r=-1]"; return; }
  if(reqNo != reqNo_recd) {
    cerr << "tts.handleDataAck: reqNo mismatch, reqNo=" << reqNo
	 << " reqNo_recd=" << reqNo_recd << endl;
    exit(99);
  }

  // parse pkt
  ServCompIFPkt_t sPkt;
  int parseRes = sPkt.parsePkt(ackPkt.data, rrBlock->reqs[reqNo].recv_len);
  if(parseRes != 1) { cerr << "parseRes = " << parseRes << endl; return; }
  if(sPkt.fixed.type != TTS_RESP) {
    cerr << "tts_cl.handleDataAck: unknown type ["
	 << ServCompIFPkt::type2str((ServCompPktType_t)sPkt.fixed.type)
	 << "]\n";
    exit(90);
  }

  handleDataAckHelper(&sPkt); // collect the data in the response
  rrBlock->freeEntry(reqNo); reqNo = -1;
} // End tts_client_info::handleDataAck()

void tts_client_info::handleDataAckHelper(ServCompIFPkt_t *sPkt) {
  const char *theText = sPkt->getValue(RESP_TEXT_F);
  if(theText == NULL) {
    cerr << "cl.handleDataAckHelper: text is NULL\n"; return;
  }

  // this is the terminating case
  if(strcmp(theText, "") == 0) {
    dbg->print() << name << " all-sentences-done p_id=" << pathID
		 << ':' << path_dest_scid << " path_ver=" << version
		 << " currSentenceReq=" << currSentenceReq << endl;
    allSentencesDone = true; return;
  }

  // sanity check
  if(currSentenceReq != aTTS_vec.size()) {
    cerr << "currSentenceReq=" << currSentenceReq << " aTTS_vec.size()="
	 << aTTS_vec.size() << endl;
    exit(77);
  }

  // store response as the next sentence
  asyncTTS_t temp_aTTS; aTTS_vec.push_back(temp_aTTS);
  asyncTTS_t &aTTS = aTTS_vec[currSentenceReq];

  strcpy(aTTS.text, theText);
  sprintf(aTTS.fName, "/tmp/tts/tts-%d-%d-%d.%d", pathID,
	  path_dest_scid, version, currSentenceReq);
  aTTS.done = aTTS.initiated = false;
  aTTS.fin = NULL;
  if(currSentenceReq == startSentence) {
    aTTS.startFilePosn = startBytePosn;
  }

  dbg->print() << name << " tts-resp p_id=" << pathID << ':'
	       << path_dest_scid << " path_ver=" << version
	       << " currSentenceReq=" << currSentenceReq << endl;

  state = PROCESSING_CL;
  if(currSentenceReq == startSentence) {
    gettimeofday(&sendStartTime, NULL); numTimeOuts = 0;
    gettimeofday(&lastSendTime, NULL);
    spawnTTSThread(currSentenceProc); // spawn the thread for the 1st sentence
    sendSoftState(); // send the first soft-state pkt

    // init the stats object
    char statFname[256];
    sprintf(statFname, "tts-%d-%d-%d.stat", pathID,
	    (int)path_dest_scid, (int)version);
    gStats.init(statFname);
  }
  currSentenceReq++;
} // End tts_client_info::handleDataAckHelper()

void tts_client_info::spawnTTSThread(int sentence_index) {
  if(sentence_index >= aTTS_vec.size()) { return; }
  asyncTTS_t &aTTS = aTTS_vec[sentence_index];
  aTTS.done = false; aTTS.initiated = true;
  dbg->print() << name << " spawn-thread p_id=" << pathID << ':'
	       << path_dest_scid << " path_ver=" << version
	       << " sentence_index=" << sentence_index << endl;
  //pthread_create(&aTTS.ttsThr, NULL, (*thr_fn), (void*)(&aTTS));
  thr_fn((void*)&aTTS); // synchronous call
} // End tts_client_info::spawnTTSThread()

void tts_client_info::sendDataReq() {
  if(upstreamPort == 0) { return; } // why am i being called?
  ServCompIFPkt_t req_sPkt;
  req_sPkt.fixed.type = TTS_REQ;
  req_sPkt.fixed.pathID = pathID;
  req_sPkt.fixed.path_dest_scid = path_dest_scid;
  req_sPkt.fixed.version = version;
  JString arg_str = long2str(currSentenceReq);
  req_sPkt.addPair(REQ_ARGS_F, arg_str.cstr());
  int send_len = req_sPkt.toStr(reqPkt.data, sizeof(reqPkt.data));
  struct timeval timeout = { 2, 0 }; // 2sec

  // send the request upstream to get text
  int send_fd = data_sock_fd; // use same socket for sending
  // send pkt through the delaying tunnel
  bool use_tunnel = (serv_my_scid != serv_upstream_scid);
  reqNo = rrBlock->insertRequest
    (send_fd, &reqPkt, send_len, upstreamIP, upstreamPort+DATA_OFFSET,
     data_ack_sock_fd, &ackPkt, timeout, use_tunnel,
     serv_upstream_scid, serv_my_scid);
  if(reqNo == -1) {
    cerr << "tts.sendDataReq: no space for request\n";
    exit(74);
  }

  dbg->print() << name << " tts-req p_id=" << pathID << ':'
	       << path_dest_scid << " path_ver=" << version
	       << " serv_my_scid=" << serv_my_scid
	       << " serv_upstream_scid=" << serv_upstream_scid
	       << " currSentenceReq=" << currSentenceReq << endl;
} // End tts_client_info::sendDataReq()

void tts_client_info::checkDataSend() {
  // see if thread has to be started
  if(currSentenceProc < aTTS_vec.size()) {
    asyncTTS_t &aTTS = aTTS_vec[currSentenceProc];
    if(!aTTS.initiated) { spawnTTSThread(currSentenceProc); }
    else {
      if(aTTS.done) {
	dbg->print() << name << " thread-done p_id=" << pathID << ':'
		     << path_dest_scid << " path_ver=" << version
		     << " sentence_index=" << currSentenceProc << endl;
	aTTS.fin = fopen(aTTS.fName, "r");
	if(aTTS.fin == NULL) { perror("fopen"); exit(78); }
	fseek(aTTS.fin, aTTS.startFilePosn, SEEK_SET);
	currSentenceProc++;
      }
    }
  }

  if(aTTS_vec.size() > 0) {
    // if something to be sent, send it
    float time_since_first_send = now(sendStartTime);
    if(time_since_first_send > (1.0*numTimeOuts*send_interval)) {
      numTimeOuts++;
      while(currSentenceSend < aTTS_vec.size()) {
	bool succ = trySend(); // send if audio file is ready
	if(succ) { break; }
      }
    }
  }

  if(reqNo == -1) {
    // check if new data req has to be sent
    if(!allSentencesDone && (currSentenceReq != startSentence) &&
       (currSentenceReq < (currSentenceSend+2))) {
      // (a) there's still some sentences left, and
      // (b) this is not the first sentence request, and
      // (c) i'm not collecting more than 2 sentences in advance
      // the reason (b) is required is because we don't want to send
      // out request before receiving request from downstream
      sendDataReq();
    }
  } else {
    // check if data req has to be resent
    int ackRecd = rrBlock->check_t_o_resend(reqNo);
    if(ackRecd == 1) {
      rrBlock->freeEntry(reqNo);
      reqNo = -1;
    } else if(ackRecd == -1) {
      rrBlock->freeEntry(reqNo);
      reqNo = -1;
      endSession();
    } else {
      // still waiting for ack
    }
  }
} // End tts_client_info::checkDataSend()

void tts_client_info::checkSoftStateSend() {
  // check if soft-state msg has to be sent
  if(state == PROCESSING_CL) {
    if(now(last_soft_state_send) > SOFT_STATE_PERIOD) {
      sendSoftState();
    }
  }
} // End tts_client_info::checkSoftStateSend()

bool tts_client_info::trySend() {
  if(currSentenceSend == aTTS_vec.size()) { return false; }
  asyncTTS_t *aTTS = &aTTS_vec[currSentenceSend];
  if(aTTS->fin == NULL) { return true; } // not yet ready

  // read data
  UDPLibPkt_t uPkt;
  myRTPpacket_t *rp = (myRTPpacket_t*)uPkt.data;
  int ret = fread((void*)rp->data, 1, send_size, aTTS->fin);
  if(ret != send_size) {
    // done with this sentence
    fclose(aTTS->fin);
    currSentenceSend++;
    startBytePosn = 0;
    return false;
  }
  startBytePosn += send_size;

  // set RTP header
  unsigned int flags = RTP_VERSION << 14 | fmt_int;
  rp->rh.rh_flags = htons(flags);
  rp->rh.rh_seqno = htons(seqno);
  seqno++;
  rp->rh.rh_ts = htonl(timestamp);
  timestamp += 160;
  rp->rh.rh_ssrc = htonl(1234567);

  int send_len = sizeof(rtphdr_t) + send_size + sizeof(uPkt) - MAXDATASZ;

  struct sockaddr_in to_addr;
  bzero((char *)&to_addr, sizeof(to_addr));
  to_addr.sin_family = PF_INET;
  to_addr.sin_addr = destIP;
  to_addr.sin_port = htons(destPort+DATA_OFFSET);

  bool use_tunnel = ((serv_my_scid != serv_dest_scid) &&
		     (serv_dest_scid != LogLine::INVALID_SCID));
  unsigned long numSentDummy;
  int res = lib_sendto(data_sock_fd, &uPkt, send_len, 0, &to_addr,
		       use_tunnel, &numSentDummy, serv_dest_scid,
		       serv_my_scid);
  if(res != send_len) {
    if(res < 0) { perror("cl.trySend.lib_sendto"); }
    fprintf(stderr, "res=%d, send_len=%d\n", res, send_len);
  }
  fprintf(stderr, "+");

  if(first_send) {
    first_send = false;
  } else {
    // statistics on gap
    float time_since_last_send = now(lastSendTime);
    float time_since_first_send = now(dbg->startTime);
    if(time_since_last_send > gapStats::GAP_THRESHOLD) {
      gStats.add(time_since_last_send, time_since_first_send);
    }
  }
  gettimeofday(&lastSendTime, NULL);

  return true;

} // End tts_client_info::trySend()

void tts_client_info::sendSoftState() {
  UDPLibPkt_t uPkt; ServCompIFPkt_t req_sPkt;
  req_sPkt.fixed.type = APP_PATH_STATE;
  req_sPkt.fixed.pathID = pathID;
  req_sPkt.fixed.path_dest_scid = path_dest_scid;
  req_sPkt.fixed.version = version;
  JString state_str =
    long2str(currSentenceSend) + ":" + long2str(startBytePosn);
  req_sPkt.addPair(PATH_STATE_F, state_str.cstr());
  int upkt_len = req_sPkt.toStr(uPkt.data, sizeof(uPkt.data));
  int send_len = upkt_len + sizeof(uPkt) - MAXDATASZ;

  struct sockaddr_in to_addr;
  bzero((char *)&to_addr, sizeof(to_addr));
  to_addr.sin_family = PF_INET;
  to_addr.sin_addr = destIP;
  to_addr.sin_port = htons(destPort+STATE_OFFSET);

  bool use_tunnel = ((serv_my_scid != serv_dest_scid) &&
		     (serv_dest_scid != LogLine::INVALID_SCID));
  unsigned long numSentDummy;
  int res = lib_sendto(state_sock_fd, &uPkt, send_len, 0, &to_addr,
		       use_tunnel, &numSentDummy, serv_dest_scid,
		       serv_my_scid);
  if(res != send_len) {
    if(res < 0) { perror("cl.sendSoftState.lib_sendto"); }
    fprintf(stderr, "res=%d, send_len=%d\n", res, send_len);
  }

  //dbg->print() << name << " send-soft-state p_id=" << pathID << ':'
  //<< path_dest_scid << " path_ver=" << version
  //<< " state_str=[" << state_str << "]" << endl;

  gettimeofday(&last_soft_state_send, NULL);
} // End tts_client_info::sendSoftState()

void tts_client_info::endSession() {
  gStats.end();
  generic_client_info::endSession();
} // End tts_client_info::endSession()
