/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef TTS_CLIENT_INFO_H
#define TTS_CLIENT_INFO_H

/** Defines info related to a client of TTS */

#include "../util/rtp.h"
#include "../app/generic_client_info.h"
#include "gapStats.h"

typedef struct asyncTTS {
  char text[4096]; // the text to be converted to speech
  char fName[256]; // the file onto which the speech should be stored
  bool done; // the flag to be set when the TTS thread finishes
  bool initiated; // indicates if a thread has been started on this
  FILE *fin; // to read the audio file
  long startFilePosn; // posn where to start streaming data
  pthread_t ttsThr; // the thread
  asyncTTS() {
    done = false; initiated = false; fin = NULL; startFilePosn = 0;
  }
} asyncTTS_t;

typedef void* (*thr_fn_t)(void *arg);

class tts_client_info: public generic_client_info {
public:
  // this is for interfacing with the tts lib -- each elt in the
  // vector represents a sentence in the email
  vector<asyncTTS_t> aTTS_vec;
  unsigned int currSentenceProc; // the current sentence being processed
  unsigned int currSentenceSend; // the current sentence being sent
  unsigned int currSentenceReq; // the current sentence being
				// requested from 'email' service
  bool allSentencesDone;
  unsigned int startSentence; // the sentence i should start with
  // the byte posn in the audio file of the first sentence, that i
  // should start sending
  long startBytePosn;

  struct timeval last_soft_state_send;
  static long SOFT_STATE_PERIOD; // in ms

  bool first_send;
  struct timeval sendStartTime; // the time when i started sending pkts
  struct timeval lastSendTime; // to measure gaps in sending
  long numTimeOuts; // the number of send_interval slots since sendStartTime
  unsigned int seqno;
  unsigned long timestamp;
  int send_interval; // in ms
  int send_size; // size of one frame (RTP payload size), in bytes
  int fmt_int; // RTP format

  static thr_fn_t thr_fn;

  gapStats_t gStats;

  tts_client_info(): generic_client_info() { init_fields(); }
  void init_fields();

  /** The thread that invokes the TTS library routines */
  void spawnTTSThread(int sentence_index);
  static void setThreadFn(thr_fn_t fn);

  void handleData();
  bool recordArgs(ServCompIFPkt_t *sPkt, JString &err_str);
  void handleDataAck();
  void handleDataAckHelper(ServCompIFPkt_t *sPkt);

  void checkDataSend();
  void checkSoftStateSend();
  /** Send if audio file is ready.  Return true if data was or can be
      sent from currSentenceSend, false otherwise */
  bool trySend();

  /** Send soft-state info downstream */
  void sendSoftState();
  /** Send request for data (upstream) -- send TTS_REQ */
  void sendDataReq();

  /** Write out gap stats also */
  void endSession();

  virtual ~tts_client_info() {}

};
typedef class tts_client_info tts_client_info_t;

#endif
