/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef REQRESPBLOCK_H
#define REQRESPBLOCK_H

#include <vector>

#include "UDPLibPkt.h"
#include "../util/newran02/include.h"
#include "../util/newran02/newran.h"

class reqInfo {
public:
  int taken; /* Has this position been taken? */
  int sending; /* Has something been sent, or are we just waiting for
		  something without having sent anything? */
  int send_fd;
  UDPLibPkt_t* send_pkt;
  int send_len; // length of the data
  struct sockaddr_in send_dest;
  int recv_fd;
  UDPLibPkt_t* recv_pkt;
  int recv_len; // length of the data (not the header)
  struct sockaddr_in recv_from;
  int recd; /* Has the response (or some other pkt) been received? */
  struct timeval timeout;
  struct timeval startTime; // When the request was issued
  int retryIndex; // index into the array of retryTimes[] -- this
  // increases upto 4 and stays there
  struct timeval lastSendTime; // time when last req was sent
  bool use_tunnel; // should i use tunnel?
  bool recd_now;
  unsigned short dest_scid; // destination's SCID
  unsigned short my_scid; // my SCID for this request
}; // End class reqInfo()

typedef class reqInfo reqInfo_t;

/** This defines the class that will be used for passing requests and
    responses between the library thread and the app threads */
class ReqRespBlock {
public:
  /** To randomize retransmit times */
  Uniform U;
  /** The times at which a UDP packet must be resent for response */
  const static long retryTimes[5]; // in ms

  vector<reqInfo_t> reqs;

  /** The total number of packets sent by this block */
  unsigned long numSent;
  /** The total number of packets recd by this block */
  unsigned long numRecd;

  /** Insert the given request into the array of requests.  Return the
      array position.  This is the entry number.  Return -1 on failure
      (have to retry after a while). */
  int insertRequest(int send_fd, UDPLibPkt_t* send_pkt, int send_len,
		    struct in_addr dest_ip, int dest_port, int recv_fd,
		    UDPLibPkt_t* recv_pkt, struct timeval timeout,
		    bool use_tunnel, unsigned short dest_scid,
		    unsigned short my_scid);

  /** Check if a response has been received for the given entry
      number.  Return 1 if response has been received, 0 if it has not
      been received yet, -1 if a timeout has occurred.  If a response
      has been received, return the length, and the from address in
      the given pointer variables (if they are not NULL) */
  int checkResp(int reqNo, int *recv_len_p, struct sockaddr_in *recv_from_p);

  /** Insert an entry for receiving a pkt.  Return the array position.
      This is the entry number.  Return -1 on failure (have to retry
      after a while). */
  int nonBlockingRecv(int recv_fd, UDPLibPkt_t* recv_pkt,
		      struct timeval timeout);

  /** Application indicates that this entry can now be freed for
      future request insertion.  Return 1 on success, 0 if error. */
  int freeEntry(int reqNo);

  /** Helper for the checkResp routine -- check if pkt has been recd */
  void check_req_helper(reqInfo_t *req);
  /** Helper for the checkResp routine -- check if pkt has to be resent */
  void check_resend(reqInfo_t *req);
  /** check for all received packets.  returns the number of packets recd */
  int check_all(int fd);
  /** check for one recd pkt, return the reqNo, -1 on failure */
  int check_one(int fd);
  /** check for timeout, and check for resend */
  int check_t_o_resend(int reqNo);

}; // End class ReqRespBlock

typedef class ReqRespBlock reqRespBlock_t;

#endif
