/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef UDP_PKT_BLOCK_H
#define UDP_PKT_BLOCK_H

#include "../util/Util.h"
#include "ReqRespBlock.h"
#include <iostream>

typedef JHashtable<char_p,int> UDPPktHash_t;

/** Class to manage memory space for a bunch of UDPLib packets */
class UDPPktBlock {
protected:
  UDPPktHash_t hash;

public:
  /** Return a UDPPktBlock for use by the application.  Returns NULL
      on failure */
  static class UDPPktBlock *getUDPPktBlock() {
    class UDPPktBlock *toRet = new UDPPktBlock();
    if(toRet != NULL) {
      toRet->hash.setHashFunction(addrHash);
    }
    return toRet;
  }

  /** Return a packet for use by the application.  Return NULL on
      failure */
  UDPLibPkt_t *getPkt() {
    return getPkt(1);
  }

  /** Same as 'getPkt()', but for this, deallocation has to happen
      'numUsers' times before this block can actually be freed.
      getPkt(1) is the same as getPkt(). */
  UDPLibPkt_t *getPkt(int numUsers) {
    UDPLibPkt_t *toRet = new UDPLibPkt_t();
    //cerr << "AllocPkt" << endl;
    if(toRet != NULL) {
      hash.put((char *)toRet, numUsers);
    }
    return toRet;
  }

  /** Set the number of users of a packet.  NOTE: getPkt() followed by
      setNumUsers(p, numUsers) is the same as getPkt(numUsers) */
  void setNumUsers(UDPLibPkt_t *p, int numUsers) {
    if(numUsers == 0) {
      hash.remove((char*)p);
      delete p;
      //cerr << "DeallocPkt" << endl;
    } else {
      hash.put((char *)p, numUsers);
    }
  }

  /** Free the given packet (allocated from here earlier) */
  void freePkt(UDPLibPkt_t *p) {
    int numToFree = hash.get((char*)p);
    numToFree--;
    if(numToFree == 0) {
      hash.remove((char*)p);
      delete p;
      //cerr << "DeallocPkt" << endl;
    } else {
      hash.put((char*)p, numToFree);
    }
  }

}; // End class UDPPktBlock

typedef class UDPPktBlock UDPPktBlock_t;

#endif
