#!/usr/bin/perl

# (c) 2002 by Bhaskaran Raman and Regents of the University of California

# Program to merge the lines of log-files from different machines

sub usage {
    print STDERR "Usage: $0 <files>\n";
} # End usage()

MAIN: {
    ($#ARGV >= 0) || do { usage(); exit(1); };

    # Open all files
    my (@LOGF);
    my $numFiles = 0;
    foreach my $file (@ARGV) {
	$LOGF[$numFiles] = "LOGF$numFiles";
	open($LOGF[$numFiles], "<$file") || die "Can't open $file: $!\n";
	print STDERR "Opened $file $LOGF[$numFiles]\n";
	$numFiles++;
    }

    # Read the first lines of each file
    my (@line);
    my (@times);
    for(my $i = 0; $i < $numFiles; $i++) {
	$line[$i] = readline($LOGF[$i]);
	if(!defined($line[$i])) {
	    $times[$i] = -1;
	} else {
	    chop($line[$i]);
	    my $dummy;
	    ($times[$i], $dummy) = split(/ /, $line[$i]);
	}
    }

    # Read all lines of all files
  READLOOP:
    for(;;) {
	# Find min of all times
	my $minIndex = 0;
	my $minExists = 0;
      MINLOOP:
	for(my $i = 0; $i < $numFiles; $i++) {
	    ($times[$i] >= 0) || do { next MINLOOP; };
	    if(($minExists == 0) || ($times[$i] < $times[$minIndex])) {
		$minIndex = $i;
		$minExists = 1;
	    }
	}

	($minExists) || do { last READLOOP; };

	# Print the min line
	print STDOUT "$line[$minIndex]\n";

	# Read another line from the min file
	$line[$minIndex] = readline($LOGF[$minIndex]);
	if(!defined($line[$minIndex])) {
	    $times[$minIndex] = -1;
	} else {
	    chop($line[$minIndex]);
	    my $dummy;
	    ($times[$minIndex], $dummy) = split(/ /, $line[$minIndex]);
	}

    } # End for(;;)

    # Close all files
    for(my $i = 0; $i < $numFiles; $i++) {
	close($LOGF[$i]);
    }

} # End MAIN
