/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Program to use UDPLib for receiving packets */

#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>

#include "UDPLib.h"
#include "../util/Util.h"

int main(int argc, char *argv[]) {
  util_init();
  int recv_fd = sock_and_bind(8866);
  if(recv_fd < 0) { perror("recv_fd socket"); exit(1); }

  UDPLib::init(); // Init the library

  reqRespBlock_t *rblock = UDPLib::getReqRespBlock(); // Get a block
  // for inserting requests
  if(rblock == NULL) {
    fprintf(stderr, "rblock is NULL\n");
    exit(1);
  }

  struct timeval t = { 5, 0 };
  UDPLibPkt_t recv_pkt;

  int reqNo =
    rblock->nonBlockingRecv(recv_fd, &recv_pkt, t);
  if(reqNo == -1) {
    fprintf(stderr, "reqNo is -1\n");
    exit(6);
  }
  fprintf(stderr, "reqNo = %d\n", reqNo);

  for(;;) {
    usleep(10000);
    int recv_len; struct sockaddr_in recv_from;
    int recd = rblock->checkResp(reqNo, &recv_len, &recv_from);
    if(recd == 0) {
      fprintf(stderr, "*");
      continue;
    }
    if(recd == 1) {
      fprintf(stderr, "+++++ pkt received from %s +++++\n",
	      inet_ntoa(recv_from.sin_addr));
      fprintf(stderr, "len = %d, data = %s\n", recv_len, recv_pkt.data);
      bool use_tunnel = false;
      unsigned long count = 0;
      UDPLib::sendResp(recv_fd, &recv_pkt, 0,
		       recv_from.sin_addr, 8888, use_tunnel, &count);
      rblock->freeEntry(reqNo);
      reqNo = rblock->nonBlockingRecv(recv_fd, &recv_pkt, t);
      if(reqNo == -1) {
	fprintf(stderr, "reqNo is -1\n");
	exit(6);
      }
      fprintf(stderr, "reqNo = %d\n", reqNo);
    }
    if(recd == -1) {
      fprintf(stderr, "----- timeout occurred -----\n");
      rblock->freeEntry(reqNo);
      reqNo = rblock->nonBlockingRecv(recv_fd, &recv_pkt, t);
      if(reqNo == -1) {
	fprintf(stderr, "reqNo is -1\n");
	exit(6);
      }
      fprintf(stderr, "reqNo = %d\n", reqNo);
    }

  } // End for(;;)

} // End main()
