/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Testing reliable request/response with UDPLib.C */

#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>

#include "UDPLib.h"
#include "../util/Util.h"

int main(int argc, char *argv[]) {
  util_init();

  int send_fd = socket(PF_INET, SOCK_DGRAM, getprotobyname("udp")->p_proto);
  if(send_fd < 0) { perror("send_fd socket"); exit(3); }

  int recv_fd = sock_and_bind(8888);

  UDPLib::init(); // Init the library

  reqRespBlock_t *rblock = UDPLib::getReqRespBlock(); // Get a block
  // for inserting requests
  if(rblock == NULL) {
    fprintf(stderr, "rblock is NULL\n");
    exit(1);
  }

  UDPLibPkt_t send_pkt;
  strcpy(send_pkt.data, "test data");
  UDPLibPkt_t recv_pkt;
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 120000;

  struct in_addr to_addr;
  inet_aton("127.0.0.1", &to_addr);
  bool use_tunnel = false;
  unsigned short pkt_dest_scid = 0; // dummy, since not using tunnel
  unsigned short my_scid = 0; // also dummy for the reason
  int reqNo =
    rblock->insertRequest(send_fd, &send_pkt, strlen(send_pkt.data)+1,
			  to_addr, 8866, recv_fd, &recv_pkt, timeout,
			  use_tunnel, pkt_dest_scid, my_scid);
  if(reqNo == -1) {
    fprintf(stderr, "reqNo is -1\n");
    exit(6);
  }
  fprintf(stderr, "reqNo = %d\n", reqNo);

  for(;;) {
    usleep(10000);
    int recd = rblock->checkResp(reqNo, NULL, NULL);
    if(recd == 0) {
      fprintf(stderr, "*");
      continue;
    }
    if(recd == 1) {
      fprintf(stderr, "+++++ response received +++++\n");
      break;
    }
    if(recd == -1) {
      fprintf(stderr, "----- timeout occurred -----\n");
      break;
    }
  }

  int recv_len;
  unsigned long count = 0, rcv_count = 0;;
  int res = UDPLib::syncReqResp
    (send_fd, &send_pkt, strlen(send_pkt.data)+1, to_addr, 8866,
     recv_fd, &recv_pkt, &recv_len, NULL, use_tunnel, &count,
     &rcv_count, pkt_dest_scid, my_scid);

  fprintf(stderr, "res = %d\n", res);

  exit(0);

} // End main()
