/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements timeSync.h */

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "timeSync.h"
#include "UDPLib.h"
#include "../util/Util.h"

char timeServer[] = "169.229.48.90";

bool getStartTimeAtHost(struct timeval *tv, struct in_addr ip) {

  if(tv == NULL) { return false; }

  int ack_listen_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(ack_listen_fd < 0) {
    perror("getTimeAtHost.socket");
    return false;
  }

  // bind to TIME_SYNC_ACK_PORT
  struct sockaddr_in listen_addr;
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(TIME_SYNC_ACK_PORT);
  if(bind(ack_listen_fd, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", TIME_SYNC_ACK_PORT);
    perror("bind failed");
    close(ack_listen_fd);
    return false;
  }

  bool use_tunnel = false;
  int send_fd = ack_listen_fd; // can use same sock for sending
  UDPLibPkt_t sendPkt, recvPkt;
  int recvLen;
  struct timeval timeout = { 2, 0 }; // 2sec

  unsigned long count = 0, rcv_count = 0;;
  unsigned short dest_scid = 0, my_scid = 0;
  int res = UDPLib::syncReqResp
    (send_fd, &sendPkt, 0, ip, TIME_SYNC_PORT, ack_listen_fd,
     &recvPkt, &recvLen, &timeout, use_tunnel, &count, &rcv_count,
     dest_scid, my_scid);

  if(res == -1) {
    fprintf(stderr, "timeout waiting for time server\n");
    close(ack_listen_fd);
    return false;
  }

  if(recvLen != 2*sizeof(struct timeval)) {
    fprintf(stderr, "invalid response from time server\n");
    close(ack_listen_fd);
    return false;
  }
  struct timeval *tvAtHost = (struct timeval*)recvPkt.data;

  struct timeval timeNowAtHost;
  struct timeval startTimeAtHost;
  timeNowAtHost.tv_sec = ntohl(tvAtHost[0].tv_sec);
  timeNowAtHost.tv_usec = ntohl(tvAtHost[0].tv_usec);
  startTimeAtHost.tv_sec = ntohl(tvAtHost[1].tv_sec);
  startTimeAtHost.tv_usec = ntohl(tvAtHost[1].tv_usec);

  unsigned long timeRelAtHost = (unsigned long)
    ((timeNowAtHost.tv_sec - startTimeAtHost.tv_sec)*1000.0 +
     (timeNowAtHost.tv_usec - startTimeAtHost.tv_usec)/1000.0);

  *tv = past(timeRelAtHost);

  close(ack_listen_fd);
  return true;

} // End getTimeAtHost()
