/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/* Implements the timeSyncServer */

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "UDPLib.h"
#include "timeSync.h"
#include "../util/Util.h"

int main(int argc, char *argv[]) {
  util_init();

  // socket
  int listen_fd = sock_and_bind(TIME_SYNC_PORT);
  if(listen_fd < 0) { perror("main.socket"); exit(1); }
  fprintf(stderr, "time server listening on port %d\a\n", TIME_SYNC_PORT);

  struct timeval startTime;
  gettimeofday(&startTime, NULL);
  time_t t = time(NULL);
  fprintf(stderr, "startTime = %s", ctime(&t));

  for(;;) {
    // get requests, and respond to them

    UDPLibPkt_t recvPkt;
    int recvLen;
    struct sockaddr_in from_addr;
    struct timeval *timeout = NULL; // wait forever
    unsigned long rcv_count = 0;
    bool pkt_ready = false;
    int res = UDPLib::blockingRecv(listen_fd, &recvPkt, &recvLen,
				   &from_addr, timeout, &rcv_count, pkt_ready);
    if(res != 1) {
      fprintf(stderr, "Error in receiving pkt\n");
      continue;
    }

    struct timeval *tv = (struct timeval*)recvPkt.data;
    struct timeval timeHere;
    gettimeofday(&timeHere, NULL);
    tv[0].tv_sec = htonl(timeHere.tv_sec);
    tv[0].tv_usec = htonl(timeHere.tv_usec);
    tv[1].tv_sec = htonl(startTime.tv_sec);
    tv[1].tv_usec = htonl(startTime.tv_usec);

    int send_fd = listen_fd; // can use same sock for sending
    bool use_tunnel = false;
    unsigned long count = 0;
    res = UDPLib::sendResp(send_fd, &recvPkt, 2*sizeof(struct timeval),
			   from_addr.sin_addr, TIME_SYNC_ACK_PORT,
			   use_tunnel, &count);

    if(res != 1) {
      fprintf(stderr, "Error sending response\n");
      continue;
    }

  } // End for(;;)

} // End main()
