/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Implements the Control Server */

#include <vector>
#include <fstream>
#include <iostream>

#include <stdio.h>
#include <signal.h>
#include <netdb.h>
#include <unistd.h>

#include "Control.h"
#include "Util.h"
#include "JString.h"

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s <scid-mapping-file>\n", progname);
} // End usage()

struct sc_node_ctl {
  unsigned short scid;
  struct in_addr ipAddr;
  bool regd;
  int conn_fd;
  sc_node_ctl() {
    regd = false;
    conn_fd = -1;
  }
};
typedef struct sc_node_ctl sc_node_ctl_t;

vector<sc_node_ctl_t> sc_list;
bool stop = false;

void intr_catcher(int) {
  if(!stop) { stop = true; }
  else { exit(0); }
} // End intr_catcher()

void readSCIDMapping(const char *scid_mapping) {
  vector<sc_node_t> sc1;
  util_readSCIDMapping(scid_mapping, sc1);
  for(unsigned int i = 0; i < sc1.size(); i++) {
    // put the mapping onto the list of sc nodes
    sc_node_ctl_t scn;
    scn.scid = sc1[i].scid;
    scn.ipAddr = sc1[i].ipAddr;
    sc_list.push_back(scn);
  }
} // End readSCIDMapping()

int main(int argc, char *argv[]) {
  util_init();
  if(argc != 2) { usage(argv[0]); exit(1); }
  const char *scid_mapping = argv[1];

  readSCIDMapping(scid_mapping);

  int s_fd = sock_and_listen(CONTROL_PORT);
  if(s_fd < 0) {
    perror("sock_and_listen"); exit(2);
  }
  cerr << "Listening for registrations on " << CONTROL_PORT << '\a' << endl;

  unsigned int numRegd = 0;
  while(numRegd < sc_list.size()) {

    // get connection from sc node
    struct sockaddr_in cli_addr;
    unsigned int clilen = sizeof(cli_addr);
    int conn_fd = accept(s_fd, (struct sockaddr *)&cli_addr, &clilen);
    if(conn_fd < 0) {
      perror("accept"); exit(4);
    }

    // which sc node did i get it from?
    int ind = -1;
    for(unsigned int i = 0; i < sc_list.size(); i++) {
      if(sc_list[i].ipAddr.s_addr == cli_addr.sin_addr.s_addr) {
	ind = i;
	break;
      }
    }

    // sanity checks
    if(ind == -1) {
      cerr << "Unknown sc node " << ip2str(cli_addr.sin_addr) << endl;
      exit(8);
    }
    if(sc_list[ind].regd) {
      cerr << "Repeat regn from scid " << sc_list[ind].scid
	   << ' ' << ip2str(sc_list[ind].ipAddr) << endl;
      exit(7);
    }

    // receive regn from sc node
    cPkt_t c2, c1;
    int res = recv(conn_fd, &c2, sizeof(c2), 0);
    if(res != sizeof(c2)) { perror("recv"); exit(5); }
    if(ntohs(c2.cmd) != REGISTER) {
      cerr << "Recv from client scid " << sc_list[ind].scid
	   << ": " << ntohs(c2.cmd) << endl;
      exit(6);
    }

    // send response
    c1.cmd = htons(REGOK);
    res = send(conn_fd, &c1, sizeof(c1), 0);
    if(res != sizeof(c1)) { perror("send"); exit(8); }

    sc_list[ind].regd = true;
    sc_list[ind].conn_fd = conn_fd;
    numRegd++;

    cerr << "Regn from " << sc_list[ind].scid << ' '
	 << ip2str(sc_list[ind].ipAddr) << endl;

  } // End while() listening for registrations

  cerr << "Sending START to all sc nodes" << endl;
  // issue start command to all sc nodes
  for(unsigned int i = 0; i < sc_list.size(); i++) {
    cPkt_t c1;
    c1.cmd = htons(START);
    int res = send(sc_list[i].conn_fd, &c1, sizeof(c1), 0);
    if(res != sizeof(c1)) { perror("send"); exit(8); }
  }

  // Activate interrupt catcher
  signal(SIGINT, intr_catcher);

  // sleep until interrupted
  for(;;) {
    if(stop) { break; }
    sleep(1);
  }

  cerr << "Sending STOP to all sc nodes" << endl;
  // issue stop command to all sc nodes
  for(unsigned int i = 0; i < sc_list.size(); i++) {
    cPkt_t c1;
    c1.cmd = htons(STOP);
    int res = send(sc_list[i].conn_fd, &c1, sizeof(c1), 0);
    if(res != sizeof(c1)) { perror("send"); exit(8); }
  }

  cerr << "Hit Ctrl-C to stop" << endl;
  for(;;) {
    sleep(5);
  }

  for(unsigned int i = 0; i < sc_list.size(); i++) {
    close(sc_list[i].conn_fd);
  }
  close(s_fd);

} // End main()
