/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
#ifndef LOGLINE_H
#define LOGLINE_H

#include <vector>
#include <netinet/in.h>

#include "JString.h"

enum desc_type {
  invalid_desc,

  ack_for_sending_e2e_patch_towards_dest_desc,
  add_path_desc,
  alt_done_desc,
  alt_path_in_progress_desc,
  am_in_repaired_sub_path_desc,
  am_last_in_sub_path_desc,
  am_not_in_posn_desc,
  am_not_in_sub_posn_desc,
  am_part_of_orig_path_desc,
  am_upstream_of_sub_path_desc,
  app_seq_not_specified_desc,

  bug_ack_from_non_peer_desc,
  bunched_pkt_too_big_desc,

  ca_timeout_desc,
  cp_done_desc,
  cp_fail_desc,
  cp_req_desc,
  cp_send_upstream_desc,
  create_alt_path_desc,
  creating_entry_for_path_desc,
  csi_req_desc,
  cs_req_desc,

  dead_links_desc,
  del_non_existent_path_desc,
  detect_time_desc,
  drop_start_desc,
  dup_hb_desc,
  dup_req_desc,
  dup_req_for_curr_version_desc,

  e2e_patch_desc,
  e2e_req_desc,
  edge_repeat_try_e2e_desc,
  err_alt_created_but_not_used_desc,
  err_incomplete_alt_path_in_use_desc,

  graph_link_dead_desc,
  graph_link_live_desc,
  graph_state_stale_desc,

  hb_ack_desc,
  hb_req_desc,

  incoming_dead_desc,
  incoming_restored_desc,
  incomplete_path_desc,
  invalid_num_in_bunch_desc,
  invalid_posn_desc,
  invalid_sub_posn_desc,

  leave_for_downstream_desc,
  local_exists_cant_handle_desc,
  local_fail_try_e2e_desc,
  lost_ack_desc,
  lost_ack_for_curr_version_desc,
  lost_ack_for_old_version_desc,
  lost_ack_for_old_version_alt_in_construction_desc,
  lost_ack_for_past_version_desc,
  ls_update_from_unknown_peer_desc,

  new_sub_path_weird_desc,
  new_sub_path_weird_try_e2e_desc,
  node_repeat_try_e2e_desc,
  non_existent_edge_del_path_desc,
  non_existent_edge_ignoring_desc,
  non_existent_path_desc,
  no_phys_path_desc,
  no_recognized_sub_path_desc,
  no_replacement_try_e2e_desc,
  nothing_to_patch_desc,
  num_in_bunch_desc,

  outgoing_dead_desc,

  parse_err_desc,
  pass_ack_desc,
  pass_req_upstream_desc,
  path_complete_desc,
  path_ends_here_desc,
  path_exists_here_desc,
  path_not_found_desc,
  paths_unequal_desc,
  pc_send_upstream_desc,
  peer_cost_change_desc,
  peer_link_dead_desc,
  peer_link_live_desc,
  phys_path_found_desc,
  pkt_drop_desc,

  rcv_fail_desc,
  removing_old_path_hb_entry_desc,
  repair_done_desc,
  repair_exists_cant_handle_desc,
  repair_exists_try_e2e_desc,
  repair_not_in_db_desc,
  replacement_found_desc,
  req_desc,

  sc_stopped_not_starting_recovery_desc,
  send_or_fwd_hb_desc,
  send_or_fwd_hb_to_unknown_peer_desc,
  send_req_upstream_desc,
  stale_to_patch_desc,
  start_desc,
  start_dead_desc,
  start_live_desc,
  sub_path_is_weird_desc,

  timing_out_path_desc,
  timing_out_sub_path_desc,
  total_recd_desc,
  total_sent_desc,
  try_local_desc,
  try_patch_desc,

  unequal_paths_desc,
  unknown_peer_desc,
  unknown_repair_path_desc,

  version_mismatch_desc
};

struct desc_type_name {
  enum desc_type type;
  char name[64];
};

enum fn_type {
  INVALID_FN,

  DEBUG_FN,

  SENDTHREAD_FN, LISTENTHREAD_FN,

  PKTDROP_FN,

  PATHEXISTS_FN,
  CHECKFAILURES_FN, PATCHPATHSFROMPEER_FN, PATCHPATHSFROMPEERRC_FN,
  CHECKPATCHQ_FN, PATCHPATH_FN, PATCHEDGELOCAL_FN,
  PATCHEDGEE2EHELPER_FN, PEERPATCHPATHLOCAL_FN, PEERPATCHBUNCH_FN,
  PEERPATCHPATHE2E_FN, SENDBUNCH_FN,

  CREATEPATH_FN, CREATESERVINST_FN, PEERCREATESERVINST_FN,
  CHANGESERVINST_FN, CHECKACKS_FN, SENDPATHHBS_FN, SENDPATHHBHELPER_FN,
  CHECKHBACKS_FN, CHECKSERVACKS_FN,

  MAIN_FN
};

struct fn_type_name {
  enum fn_type type;
  char name[64];
};

struct patch_sub_s {
  short num_to_patch, num_in_bunch;
  short totLen;
  void init() {
    num_to_patch = num_in_bunch = -2;
    totLen = -2;
  }
};

struct ls_sub_s {
  short currRTT, lastRTT;
  void init() { currRTT = lastRTT = -2; }
};

struct emul_sub_s {
  long seedLong;
  long gapInFuture;
  void init() { seedLong = -2; gapInFuture = -2; }
};

struct err_sub_s {
  short dup;
  struct in_addr peerIP;
  unsigned short curr_path_ver, curr_alt_ver;
  short lastHBPosn;
  void init() {
    dup = -2; peerIP.s_addr = (unsigned long)-2;
    curr_path_ver = curr_alt_ver = 0;
    lastHBPosn = -2;
  }
};

struct serv_sub_s {
  struct in_addr sIP;
  int sPort;
  void init() { sIP.s_addr = (unsigned long)-2; sPort = -2; }
};

enum u_type { INVALID_U, PATCH_U, LS_U, EMUL_U, ERR_U, SERV_U };

enum log_key {
  INVALID_LK,
  FILE_LK,
  NEWSUBPATH_LK,
  DEAD_LK,
  SERVICES_LK, PHYSPATH_LK,
  SNAME_LK
};

struct Log_KeyValuePair_LK {
  enum log_key key;
  JString value;
}; // End struct Log_KeyValuePair_LK
typedef struct Log_KeyValuePair_LK Log_KeyValuePair_LK_t;

struct Log_KeyValuePair {
  JString key;
  JString value;
}; // End struct Log_KeyValuePair
typedef struct Log_KeyValuePair Log_KeyValuePair_t;

/** Represents a line in the merged log-file */
struct LogLine {
  static unsigned short INVALID_SCID;
  static JString EMPTY_STR;

  static struct fn_type_name fn_type_name_map[];
  static enum fn_type get_fn_type(const JString &name);
  static const char *get_fn_name(enum fn_type type);

  static struct desc_type_name desc_type_name_map[];
  static enum desc_type get_desc_type(const JString &name);
  static const char *get_desc_name(enum desc_type type);

  bool succ;
  bool read_mode;

  unsigned long time;
  unsigned short scid;
  enum fn_type fn;
  enum desc_type desc;
  vector<Log_KeyValuePair_t> kv;
  vector<Log_KeyValuePair_LK_t> kv_lk;

  unsigned short dest_scid, origin_scid;
  unsigned short nextSCID, peerSCID;
  int pathID; unsigned short path_dest_scid;
  unsigned short version, fail_version;
  unsigned short path_origin_scid;
  unsigned short fail_dest_scid, fail_origin_scid;
  short failPosn;

  short posn, subPosn;

  unsigned long total_sent, total_recd;

  int success;
  enum u_type utype;

  union {
    struct patch_sub_s patch;
    struct ls_sub_s ls;
    struct emul_sub_s emul;
    struct err_sub_s err;
    struct serv_sub_s serv;
  } u;

  LogLine(); // init all values
  LogLine(const char *line); // parse from given line
  void init(); // init all values
  bool parse(const char *line); // parse from given line
  void err(const char *msg); // internal function
  JString toStr(); // convert back to JString
  JString printValid(); // print only the valid fields
  const JString &getValue(const JString &key) const; // get val corr to JString
  JString strip_getValue(const JString &key) const; // get val and strip []
  void addPair(enum log_key key, const JString &value);
}; // End class LogLine
typedef class LogLine LogLine_t;

#endif
