/** (c) 2002 by Bhaskaran Raman and Regents of the University of California */
/** Set of utility routines */

#include <stdio.h>
#include <fstream>
#include <unistd.h>
#include <netdb.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if.h>

#include "Util.h"

char TUN_HOST[] = "169.229.48.162";
struct in_addr tun_ip;
int TUN_PORT = 8866;

bool gdbg = false;
bool cfg_to_file = false;
bool cfg_use_stream = false;
bool cfg_read_paths = true;
bool cfg_opt_logline = false;
struct in_addr myIP;

void util_init() {
  int ip_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  struct ifconf ifs;
  struct ifreq ifr[5];
  ifs.ifc_len = 5*sizeof(struct ifreq);
  ifs.ifc_req = ifr;
  int res = ioctl(ip_fd, SIOCGIFCONF, &ifs);
  if(res != 0) { perror("Error in ioctl call"); exit(3); }

  for(unsigned int i = 0; i < ifs.ifc_len/sizeof(struct ifreq); i++) {
    if(ioctl(ip_fd, SIOCGIFFLAGS, &ifr[i]) < 0) {
      continue;
    }

    /*
    fprintf(stderr, "%s %d %d\n", ifr[i].ifr_name,
	    (int)(ifr[i].ifr_flags & IFF_LOOPBACK),
	    (int)(ifr[i].ifr_flags & IFF_UP));
    */

    if((!(ifr[i].ifr_flags & IFF_LOOPBACK)) &&
       (ifr[i].ifr_flags & IFF_UP)) {
      if(ioctl(ip_fd, SIOCGIFADDR, &ifr[i]) < 0) {
	continue;
      }
      myIP = (*((struct sockaddr_in *)&ifr[i].ifr_addr)).sin_addr;

      if(ioctl(ip_fd, SIOCGIFINDEX, &ifr[i]) < 0) {
	continue;
      }
      break;
    }
  }
  close(ip_fd);

  /*
  const char *hostname = getenv("HOSTNAME");
  if(hostname == NULL) { fprintf(stderr, "can't get HOSTNAME\n"); exit(4); }
  myIP = nslookup(hostname);
  */
  tun_ip.s_addr = inet_addr(TUN_HOST);
} // End util_init()

struct in_addr nslookup(const char *host) {
  struct hostent *hp = gethostbyname(host);
  if(hp == 0) {
    fprintf(stderr, "could not abtain address of %s\n", host);
    exit(250);
  }
  struct in_addr toRet;
  bcopy(hp->h_addr_list[0], (caddr_t)&toRet, hp->h_length);
  return toRet;
} // End nslookup()

int sock_and_bind(int port) {
  // socket for listening
  int listen_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(listen_fd < 0) {
    perror("sock_and_bind");
    return listen_fd;
  }

  // bind to SERV_COMP_PORT
  struct sockaddr_in listen_addr;
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(port);
  if(bind(listen_fd, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", port);
    perror("bind failed");
    close(listen_fd);
    return -1;
  }
  return listen_fd;
} // End sock_and_bind()

int sock_and_connect(const char *host, int port) {
  int sock_fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(sock_fd < 0) { perror("socket failed"); return -1; }
  struct sockaddr_in serv_addr;

  /* connect */
  serv_addr.sin_addr = nslookup(host);
  serv_addr.sin_family = PF_INET;
  serv_addr.sin_port = htons(port);
  int res = connect(sock_fd, (struct sockaddr *)&serv_addr,
		    sizeof(serv_addr));
  if(res != 0) { perror("connect failed"); close(sock_fd); return -1; }
  return sock_fd;
} // End sock_and_connect()

int sock_and_listen(int port) {
  int sock_fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(sock_fd < 0) { perror("socket failed"); return -1; }
  struct sockaddr_in serv_addr;

  /* bind, listen */
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = INADDR_ANY;
  serv_addr.sin_port = htons(port);
  if(bind(sock_fd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    perror("bind"); close(sock_fd); return -1;
  }

  if(listen(sock_fd, SOMAXCONN) < 0) {
    perror("listen"); close(sock_fd); return -1;
  }
  return sock_fd;
} // End sock_and_listen()

int lib_sendto(int fd, UDPLibPkt_t *upkt, unsigned int len, int flags,
	       struct sockaddr_in *to_addr, bool use_tunnel,
	       unsigned long *count, unsigned short dest_scid,
	       unsigned short my_scid) {
#ifdef IP_TUNNEL
  // save the old addresses in the pkt itself (the tunnel forwarder
  // will look at these)
  upkt->dest_addr.s_addr = to_addr->sin_addr.s_addr;
  upkt->dest_port = to_addr->sin_port;
  upkt->dest_scid = htons(dest_scid); upkt->my_scid = htons(my_scid);

  if(!use_tunnel) {
    return sendto(fd, (void*)upkt, len, flags, (struct sockaddr*)to_addr,
		  sizeof(struct sockaddr_in));
  }

  // have to send the pkt through tunnel

  // form the new address
  struct sockaddr_in tun_to_addr;
  bzero((char *)&tun_to_addr, sizeof(tun_to_addr));
  tun_to_addr.sin_family = AF_INET;
  tun_to_addr.sin_addr.s_addr = tun_ip.s_addr;
  tun_to_addr.sin_port = htons(TUN_PORT);

  // send the pkt now
  (*count)++;
  return sendto(fd, (void*)upkt, len, flags,
		(struct sockaddr*)&tun_to_addr, sizeof(tun_to_addr));

#else
  /*
  if((ntohs(to_addr->sin_port) >= 3500) &&
     (ntohs(to_addr->sin_port) <= 3504)) {
    fprintf(stderr, "lib_sendto: port=%d, len=%d, to_addr=%s\n",
	    ntohs(to_addr->sin_port), len, inet_ntoa(to_addr->sin_addr));
  }
  */
  return sendto(fd, (void*)upkt, len, flags, (struct sockaddr*)to_addr,
		sizeof(struct sockaddr_in));

#endif
} // End lib_sendto()

int lib_recvfrom(int fd, UDPLibPkt_t *upkt, unsigned int len,
		 struct sockaddr_in *from_addr) {
  unsigned int from_len = sizeof(struct sockaddr_in);
  int res = recvfrom(fd, (void*)upkt, len, 0,
		     (struct sockaddr*)from_addr, &from_len);
#ifdef IP_TUNNEL
  if(ntohs(upkt->src_dest_same)) {
    from_addr->sin_addr.s_addr = myIP.s_addr;
  }
#endif
  return res;
} // End lib_recvfrom()

float time2float(struct timeval t) {
  return t.tv_sec*1000.0 + t.tv_usec/1000.0;
} // End time2float()

float now(struct timeval start) {
  struct timeval t;
  gettimeofday(&t, NULL);
  return (t.tv_sec - start.tv_sec)*1000.0 +
    (t.tv_usec - start.tv_usec)/1000.0;
} // End now()

struct timeval past(unsigned long t) {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  unsigned long t_sec = t/1000;
  unsigned long t_micros = (t%1000)*1000;
  tv.tv_sec -= t_sec;
  if(tv.tv_usec >= t_micros) {
    tv.tv_usec -= t_micros;
  } else {
    tv.tv_sec--;
    tv.tv_usec += (1000*1000 - t_micros);
  }
  return tv;
} // End past()

long int timestamp() {
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec;
} // End timestamp()

unsigned int ulongHash(const unsigned long &l) {
  register unsigned long key = l;
  // 2654435761 is the golden ratio of 2^32. The right shift of 3 bits
  // assumes that all objects are aligned on the 8 byte boundary. If a
  // system aligns on the 4 byte boundary, then a right shift of 2
  // bits should be done.
  return (key >> 3) * 2654435761U;
} // End ulongHash()

unsigned int ushortHash(const unsigned short &u) {
  return u;
} // End ushortHash()

unsigned int addrHash(const char_p &addr) {
  register unsigned long key;
  key = (unsigned long) addr;
  // 2654435761 is the golden ratio of 2^32. The right shift of 3 bits
  // assumes that all objects are aligned on the 8 byte boundary. If a
  // system aligns on the 4 byte boundary, then a right shift of 2
  // bits should be done.
  return (key >> 3) * 2654435761U;
} // End addrHash()

unsigned int JStringHash(const JString &s) {
  unsigned int h=0,g;
  int i; int len = s.length();
  for(i=0;i<len;i++) {
        h=(h<<4)+(s[i]);
        if((g=(h&0xf0000000))) {
          h=h^(g>>24);
          h=h^g;
        }
  }
  return h;
} // End JStringHash()

char *my_strcpy(char *dest, const char *src) {
  int i = 0;
  while(src[i]) { dest[i] = src[i]; i++; }
  dest[i] = 0;
  return (dest+i);
} // End my_strcpy()

char *longPrint(char *s, const long argval) {
  int slen = 0;
  long val = argval;
  if(argval < 0) { val = -argval; }
  for(;;) {
    s[slen] = '0' + (val%10);
    slen++;
    val = val/10;
    if(val == 0) break;
  }
  if(argval < 0) { s[slen] = '-'; slen++; }
  // reverse JString
  for(int i = 0; i < slen/2; i++) {
    char temp = s[i];
    s[i] = s[slen-i-1];
    s[slen-i-1] = temp;
  }
  s[slen] = 0;
  return (s+slen);
} // End longPrint()

JString long2str(const long argval) {
  char s[64]; longPrint(s, argval);
  return JString(s);
} // End long2str()

char *ipPrint(char *s, const struct in_addr ip) {
  unsigned long ipLong = ntohl(ip.s_addr);
  unsigned long fourth8 = ipLong & 0x000000ff;
  ipLong = ipLong >> 8;
  unsigned long third8 = ipLong & 0x000000ff;
  ipLong = ipLong >> 8;
  unsigned long second8 = ipLong & 0x000000ff;
  ipLong = ipLong >> 8;
  unsigned long first8 = ipLong & 0x000000ff;
  char *lastPos = s;
  lastPos = longPrint(lastPos, first8); *lastPos = '.'; lastPos++;
  lastPos = longPrint(lastPos, second8); *lastPos = '.'; lastPos++;
  lastPos = longPrint(lastPos, third8); *lastPos = '.'; lastPos++;
  lastPos = longPrint(lastPos, fourth8); *lastPos = 0; lastPos++;
  return lastPos;
} // End ipPrint()

JString ip2str(const struct in_addr ip) {
  char s[128]; ipPrint(s, ip);
  return JString(s);
} // End ip2str()

int JString_atoi(const JString &s) {
  return atoi(s.cstr());
} // End JString_atoi()

int str_split(const char *a_s, vector<JString> &vec) {
  int pos = 0;
  int numPieces = 0;
  char s[4096]; strcpy(s, a_s);
  while(s[pos]) {
    if(isspace(s[pos])) {
      pos++; continue;
    } else {
      int startPos = pos;
      while((s[pos]) && !isspace(s[pos])) {
	pos++;
      }
      char temp = s[pos]; s[pos] = 0;
      vec.push_back(JString(&s[startPos]));
      s[pos] = temp;
      numPieces++;
    }
  }
  return numPieces;
} // End str_split()

int str_split_mutate(char *s, vector<char*> &vec) {
  int pos = 0;
  int numPieces = 0;
  while(s[pos]) {
    if(isspace(s[pos])) {
      pos++; continue;
    } else {
      int startPos = pos;
      while((s[pos]) && !isspace(s[pos])) {
	pos++;
      }
      vec.push_back(&s[startPos]);
      numPieces++;
      if(s[pos] == 0) { break; }
      else { s[pos] = 0; pos++; }
    }
  }
  return numPieces;
} // End str_split_mutate()

int split(const JString &s, vector<JString> &vec) {
  int len = s.length();
  int pos = 0;
  int numPieces = 0;
  while(pos < len) {
    if(isspace(s[pos])) {
      pos++; continue;
    } else {
      int startPos = pos;
      while((pos < len) && !isspace(s[pos])) {
	pos++;
      }
      int subLen = pos-startPos;
      vec.push_back(s.substr(startPos, subLen));
      numPieces++;
    }
  }
  return numPieces;
} // End split()

JString join(const vector<JString> &sVec) {
  JString newJString = "";
  for(unsigned int i = 0; i < sVec.size(); i++) {
    newJString += (sVec[i] + " ");
  }
  return newJString;
} // End join()

void JString_cp(char *buf, const JString &s) {
  for(unsigned int j = 0; j < s.length(); j++) {
    buf[j] = s[j];
  }
  buf[s.length()] = 0;
} // End JString_cp()

void util_readSCIDMapping(const char *scid_mapping,
			  vector<sc_node_t> &sc_list) {
  cerr << "reading scid_mapping file " << scid_mapping << endl;
  ifstream p_fin(scid_mapping);
  for(;;) {
    // read each line
    if(p_fin.eof()) break;
    char line[256];
    p_fin.getline(line, sizeof(line)); // read a line
    if(p_fin.eof()) break;
    if(p_fin.fail()) {
      cerr << "Error reading " << scid_mapping << endl;
      exit(99);
    }

    // line has the format: "scid  hostName"
    vector<JString> sVec;
    int numFields = split(line, sVec);
    if(numFields != 2) {
      cerr << "Format of " << scid_mapping << " is wrong" << endl;
      exit(97);
    }

    // get the scid value
    int scidInt = JString_atoi(sVec[0]);
    if(scidInt <= 0) {
      cerr << "scid is " << scidInt << " (has to be > 0)" << endl;
      exit(98);
    }

    // put the mapping onto the list of sc nodes
    sc_node_t scn;
    scn.scid = (unsigned short)scidInt;
    scn.ipAddr = nslookup(sVec[1].cstr()); // DNS lookup
    sc_list.push_back(scn);

  } // End for(;;) reading all lines in config file
  p_fin.close();
} // End readSCIDMapping()
